/* 
 * ShapeView.java 1.1 2002-10-18
 *
 *  2001-2002 Jouni Smed
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of the University of Turku nor the names of its
 *       contributors may be used to endorse or promote products derived 
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package fi.utu.cs.appearance;

import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.geom.AffineTransform;
import java.awt.geom.Rectangle2D;
import java.awt.Shape;
import java.awt.Dimension;
import javax.swing.JPanel;
import java.util.List;
import java.util.ListIterator;

/**
 *  A Swing component that displays a list of <code>Shapable</code>
 *  objects. The list determines the sequence in which the objects
 *  are rendered (i.e., the subsequent objects are rendered over
 *  the currently rendered ones). If the object is an instance of
 *  the class <code>MultiShapable</code>, the shapable object given
 *  by its iterator are rendered first followed by the shapable
 *  object itself. The view is associated with a zoom 
 *  factor by which the coordinates are scaled when transformed to the 
 *  screen. The coordinate system can be either traditional screen 
 *  coordinate system, where the y axis points down, or normal 
 *  coordinate system, where the y axis points up.
 *
 *  @see Shapable
 *	@see MultiShapable
 *  @version 1.1, 021018
 *  @author Jouni Smed 
 */

public class ShapeView extends JPanel 
  {
  /** A list comprising <code>Shapable</code> objects. */
  private List shapes;
  /** An affine transform matrix for the view. */
  private AffineTransform transform;
  /** A zoom factor the view. */
  private double zoomFactor = 0.0;
  /** Indicates the used coordinate system. */
  private boolean screenCoordinateSystem;
  /** A rectangle defining the visible region of the view. */ 
  private Rectangle2D visibleRegion = null;
  /**
   *	???????????
   *	@since 1.1
   */
  private boolean rotated;
  
  /** 
   *  Creates a view. The zoom factor is 1.0 and the view uses 
   *  the screen coordinate system.
   */
  public ShapeView()
    {
    this(null, 1.0, true);
    }

  /** 
   *  Creates a view to the shapable list <code>s</code>. 
   *  The zoom factor is 1.0 and the view uses the screen coordinate
   *  system.
   */
  public ShapeView(List s)
    {
    this(s, 1.0, true);
    }

  /** 
   * Creates a view to the shapable list <code>s</code>
   * with the zoom factor <code>z</code>. The view uses the
   * screen coordinate system.
   */
  public ShapeView(List s, double z)
    {
    this(s, z, true);
    }

  /** 
   * Creates a view to the shapable list <code>s</code>
   * with the zoom factor <code>z</code> using the screen coordinate
   * system if <code>c == true</code> or the normal coordinate
   * system if <code>c == false</code>. 
   */
  public ShapeView(List s, double z, boolean c)
    {
    super();
    transform = new AffineTransform();
    screenCoordinateSystem = c;
    setZoomFactor(z);
    setShapes(s);
    rotated = false;
    }

	public void setRotated()
		{
		rotated = true;
		}


  /** Returns the zoom factor associated with this view. */
  public double getZoomFactor()
    {
    return zoomFactor;
    }

  /** Sets the zoom factor to <code>z</code>. */
  public void setZoomFactor(double z)
    {
    if (zoomFactor != 0.0)
    	{
	    double oz = 1.0 / zoomFactor;
	    if (screenCoordinateSystem)
	      {
	      transform.scale(oz, oz);
	      }
	    else 
	      {
	      transform.scale(oz, -oz);
	      }    
	    }
    zoomFactor = z;
    if (screenCoordinateSystem)
      {
      transform.scale(z, z);
      }
    else 
      {
      transform.scale(z, -z);
      }
    if (visibleRegion != null) updateSize();
    }

  /** Sets the size according to the visible region and zoom factor. */
  private void updateSize()
    {
    int width = (int)((visibleRegion.getMaxX() - visibleRegion.getMinX()) 
                * zoomFactor);
    int height = (int)((visibleRegion.getMaxY() - visibleRegion.getMinY()) 
                * zoomFactor);
    if (rotated)
    	{
    	int temp = width;
    	width = height;
    	height = temp;
    	}
    Dimension size = new Dimension(width, height);
    setMinimumSize(size);
    setMaximumSize(size);
    setPreferredSize(size);
    revalidate();
    repaint();
    }

  /** 
   *  Returns <code>true</code> if the view uses the screen coordinate
   *  system.
   *  @see #setScreenCoordinateSystem()
   *  @see #setNormalCoordinateSystem()
   */ 
  public boolean isScreenCoordinateSystem()
    {
    return screenCoordinateSystem;
    }

  /** 
   *  Returns <code>true</code> if the view uses the normal coordinate
   *  system.
   *  @see #setScreenCoordinateSystem()
   *  @see #setNormalCoordinateSystem()
   */ 
  public boolean isNormalCoordinateSystem()
    {
    return !screenCoordinateSystem;
    }

  /** Sets the screen coordinate system to this view. */
  public void setScreenCoordinateSystem()
    {
    screenCoordinateSystem = true;
    setZoomFactor(zoomFactor);    
    }

  /** Sets the normal coordinate system to this view. */
  public void setNormalCoordinateSystem()
    {
    screenCoordinateSystem = false;
    setZoomFactor(zoomFactor);    
    }

  /** Sets the visible region to the rectangle <code>r</code>.*/
  public void setVisibleRegion(Rectangle2D r)
    {
    visibleRegion = r;
    updateSize();
    }

  /** Sets the shapable list to <code>s</code>. */
  public void setShapes(List s)
    {
    shapes = s;
    }

	/** Renders a shapable object. */
	private void render(Graphics2D g2, Shapable s)
		{
    if (s.isFilled())
      {
      g2.setPaint(s.paint());
      g2.fill(s.shape());
     	}
    if (s.isDrawn())
      {
      g2.setPaint(s.color());
      g2.setStroke(s.stroke());
      g2.draw(s.shape());
      }
		}


  /** 
   *  Draws the view when invoked by Swing. Applications should not invoke 
   *  this method directly, but should instead use the inherited repaint 
   *  method to schedule the view for redrawing. 
   *  @see #repaint()
   */
  public void paintComponent(Graphics g)
    {
    super.paintComponent(g);
    render(g);
    }  

	/**
	 *	Renders the view. This methods should be used only when the 
	 *	system supported repainting is not available (e.g., in full
	 *	screen mode). Other uses for this method are strongly 
	 *	discouraged; instead, call the <code>repaint()</code> method.
	 *
   *  @see #repaint()
	 *	@since 1.1
	 */
	public void render(Graphics g)
		{
    if (shapes == null) return;
    Graphics2D g2 = (Graphics2D)g;
    AffineTransform af = new AffineTransform(transform);
    if (rotated)
    	{
			af.rotate(-Math.PI / 2.0);
			af.translate(0.0, 60.0);
    	}
    if (!screenCoordinateSystem)
      {
      af.translate(-visibleRegion.getMinX(), -visibleRegion.getMaxY());
      }
    else 
      {
      af.translate(-visibleRegion.getMinX(), -visibleRegion.getMinY());     
      }
    g2.setTransform(af);
    g2.setClip(visibleRegion);
    ListIterator it = shapes.listIterator();
    while (it.hasNext())
      {
      Shapable s = (Shapable)it.next();
      if (s instanceof MultiShapable)
      	{
      	MultiShapable ms = (MultiShapable)s;
      	ShapeIterator is = ms.iterator();
      	while (is.hasNext())
      		render(g2, is.next());
      	}
      render(g2, s);
      } 
		}
  } 
