/* 
 * AIsHockeyClient.java 1.0 2002-08-29
 *
 *  2002 Jouni Smed
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of the University of Turku nor the names of its
 *       contributors may be used to endorse or promote products derived 
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package fi.utu.cs.hockey;

import fi.utu.cs.hockey.ai.AI;
import fi.utu.cs.hockey.Position;

import java.util.*;
import java.io.*;
import java.awt.*;

/**
 *  A hockey game client. 
 *
 *  @version 1.0, 020829
 *  @author  Jouni Smed
 */
public class AIsHockeyClient 
	{
	private static final String[] 
					 keywords = { "SERVER_ADDRESS", "SERVER_PORT", "INIT_PORT",
					 							"GROUP_ADDRESS", "GROUP_PORT",
					 							"TEAM", "HELMET", "JERSEY", "GOALKEEPER",
											  "LEFT_DEFENSE", "RIGHT_DEFENSE", "LEFT_WING",
											  "RIGHT_WING", "CENTER", "NEVER_QUIT" };
											  
	private static final int TAG_UNKNOWN = -1;
	private static final int TAG_SERVER_ADDRESS = 0;
	private static final int TAG_SERVER_PORT = 1;
	private static final int TAG_INIT_PORT = 2;
	private static final int TAG_GROUP_ADDRESS = 3;
	private static final int TAG_GROUP_PORT = 4;
	private static final int TAG_TEAM = 5;
	private static final int TAG_HELMET = 6;
	private static final int TAG_JERSEY = 7;
	private static final int TAG_GOALKEEPER = 8;
	private static final int TAG_LEFT_DEFENSE = 9;
	private static final int TAG_RIGHT_DEFENSE = 10;
	private static final int TAG_LEFT_WING = 11;
	private static final int TAG_RIGHT_WING = 12;
	private static final int TAG_CENTER = 13;
	private static final int TAG_NEVER_QUIT = 14;

	private String serverAddress = "";
	private int serverPort = 0;
	private int initPort = 0;
	private String groupAddress = "";
	private int groupPort = 0;
	private String teamName = "";
	private Color helmetColor = Color.blue;
	private Color jerseyColor = Color.white;
	private AI[] ais = new AI[6];
	private boolean neverQuit = false;
	
	/** 
	 *	Constructs a hockey client based on the initialization data 
	 *  from the stream <code>in</code>.
	 */
	protected AIsHockeyClient(InputStream in)
		{
		BufferedReader reader = new BufferedReader(new InputStreamReader(in));
		String row;		
		try
			 {
			 row = reader.readLine();
			 while (row != null)
					{
					int index = TAG_UNKNOWN;
					for (int i = 0; i < keywords.length; i++)
						{
						if (row.startsWith(keywords[i]))
							{
							index = i;
							break;
							}
						}
					if (index != TAG_UNKNOWN)
						{
						String argument = row.substring(keywords[index].length());
						argument = argument.trim();
						switch (index) 
							{
							case TAG_SERVER_ADDRESS:
								serverAddress = argument;
								break;
							case TAG_SERVER_PORT:
								try
									{
									serverPort = Integer.parseInt(argument);
									}
								catch (NumberFormatException e)
									{
									System.err.println("Not a number: " + argument);
									System.exit(1);
									}
								break;
							case TAG_INIT_PORT:
								try
									{
									initPort = Integer.parseInt(argument);
									}
								catch (NumberFormatException e)
									{
									System.err.println("Not a number: " + argument);
									System.exit(1);
									}
								break;
							case TAG_GROUP_ADDRESS:
								groupAddress = argument;
								break;
							case TAG_GROUP_PORT:
								try
									{
									groupPort = Integer.parseInt(argument);
									}
								catch (NumberFormatException e)
									{
									System.err.println("Not a number: " + argument);
									System.exit(1);
									}
								break;
							case TAG_TEAM:
								teamName = argument;
								break;
							case TAG_HELMET:
								try
									{
									helmetColor = Color.decode(argument);
									}
								catch (NumberFormatException e)
									{
									System.err.println("Not a color: " + argument);
									System.exit(1);
									}
								break;
							case TAG_JERSEY:
								try
									{
									jerseyColor = Color.decode(argument);
									}
								catch (NumberFormatException e)
									{
									System.err.println("Not a color: " + argument);
									System.exit(1);
									}
								break;
							case TAG_NEVER_QUIT:
								neverQuit = true;
								break;
							default:
								StringTokenizer st = new StringTokenizer(argument);
								String aiName = null;
								String number = null;
								String name = null;
								if (st.hasMoreTokens()) aiName = st.nextToken();
								if (st.hasMoreTokens()) number = st.nextToken();
								while (st.hasMoreTokens()) 
									{
									if (name == null) name = st.nextToken();
									else name = name + " " + st.nextToken();
									}
								if (aiName == null || number == null || name == null)
									{
									System.err.println("Row malformed: " + row);
									System.exit(1);
									}
							
								AI ai = null;
								ClassLoader loader = new AILoader();
								try
									{
									Class c = loader.loadClass(aiName);
									ai = (AI)c.newInstance();
									}
								catch (ClassNotFoundException e)
									{
									System.err.println("Cannot find AI class: " + aiName);
									System.exit(1);
									}
								catch (InstantiationException e)
									{
									System.err.println("Cannot instantiate AI class: " + aiName);
									System.exit(1);
									}
								catch (IllegalAccessException e)
									{
									System.err.println("Cannot access AI class: " + aiName);
									System.exit(1);
									}
								ai.setPlayerName(name);
								try
									{
									ai.setNumber(Integer.parseInt(number));
									}
								catch (NumberFormatException e)
									{
									System.err.println("Not a number: " + number);
									System.exit(1);
									}
								switch (index)
									{
									case TAG_GOALKEEPER:
										ai.setPosition(Position.GOALKEEPER);
										ais[Position.GOALKEEPER] = ai;
										break;
									case TAG_LEFT_DEFENSE:
										ai.setPosition(Position.LEFT_DEFENSE);
										ais[Position.LEFT_DEFENSE] = ai;
										break;
									case TAG_RIGHT_DEFENSE:
										ai.setPosition(Position.RIGHT_DEFENSE);
										ais[Position.RIGHT_DEFENSE] = ai;
										break;
									case TAG_LEFT_WING:
										ai.setPosition(Position.LEFT_WING);
										ais[Position.LEFT_WING] = ai;
										break;
									case TAG_RIGHT_WING:
										ai.setPosition(Position.RIGHT_WING);
										ais[Position.RIGHT_WING] = ai;
										break;
									case TAG_CENTER:
										ai.setPosition(Position.CENTER);
										ais[Position.CENTER] = ai;
										break;
									}
							}						
						}
					row = reader.readLine();
					}
			 reader.close();
			 }
		catch (IOException e)
			 {
			 System.err.println("Configuration parse failed.");
			 System.exit(1);   
			 }

		System.out.println("Connecting");
		System.out.println("  Server: " + serverAddress);
		System.out.println("    Port: " + serverPort);
		System.out.println("    Init: " + initPort);
		System.out.println("  Group:  " + groupAddress);
		System.out.println("    Port: " + groupPort);
		System.out.println();
		System.out.println("Team: " + teamName);
		for (int i = 0; i < ais.length; i++)
			if (ais[i] != null)
				{
				if (neverQuit) ais[i].setNeverQuit();
				ais[i].setTeamName(teamName);
				ais[i].setHelmetColor(helmetColor);
				ais[i].setJerseyColor(jerseyColor);
				ais[i].connect(serverAddress, serverPort, initPort,
											 groupAddress, groupPort);
				new Thread(ais[i]).start();
				System.out.println("  Player #" + ais[i].getNumber() 
													 + " " + ais[i].getPlayerName());
				}
		}



	private class AILoader extends ClassLoader
		{
		protected Class findClass(String name) throws ClassNotFoundException
			{
			Class newClass = findLoadedClass(name);
			if (newClass == null)
				{
					newClass = findSystemClass(name);
					if (newClass != null) return newClass;
				}
			return newClass;
			}
		}		
		
	}