/* 
 * Game.java 1.0 2002-11-25
 *
 *  2002 Jouni Smed
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of the University of Turku nor the names of its
 *       contributors may be used to endorse or promote products derived 
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package fi.utu.cs.hockey;

import fi.utu.cs.appearance.*;
import fi.utu.cs.physics.*;
import fi.utu.cs.hockey.rink.*;
import fi.utu.cs.hockey.rink.FaceOff;
import fi.utu.cs.hockey.net.Server;

import java.awt.Color;
import java.util.Random;
import java.util.LinkedList;
import java.util.Observer;
import java.util.Observable;
import java.awt.Toolkit;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.awt.geom.Rectangle2D;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.geom.AffineTransform;
import java.awt.image.BufferStrategy;
import javax.swing.JLabel;
import java.applet.Applet;
import java.applet.AudioClip;
import java.net.URL;
import java.net.MalformedURLException;


/**
 *  A hockey game. 
 *
 *  @version 1.0, 021017
 *  @author  Jouni Smed
 */

public class Game extends Thread
  {
  private static final int INIT_TIME_SLICE_MS = 50;
  private static final int FAST_TIME_SLICE_MS = 5;
  private static final double INIT_FACEOFF_DELAY = 2000;
  private static final int GAME_OVER_REPEATS = 200;
  private static final int GAME_OVER_DELAY = 5;

  private static final Point2 PUCK_SAFE_POSITION = 
            new Point2(-30.0, 0.0);
  
  private static final double MAX_PUCK_NOISE = 0.05;

  private static final String[] POSITION_SHORTCUT = {"G", "LD", "RD", 
                                                    "LW", "RW", "C" };
  private ShapeView view;
  private Universe u;
  private static final double INIT_ZOOM = 10.0;
  
  private Random random = new Random();

  private Puck puck; 
  private Rink rink;
  private int scoreHome = 0;
  private int scoreVisitor = 0;
  private Team teamHome;
  private Team teamVisitor;

  private long time;
  private JLabel gameStatusTime = null;
  private JLabel gameStatusScore = null;
  private JLabel gameStatusRoster = null;
  private JLabel gameStatusReferee = null;
  
  private int timeSliceMs = INIT_TIME_SLICE_MS;
  private double timeSliceS = INIT_TIME_SLICE_MS / 1000.0;
  private double zoom = INIT_ZOOM;

  private boolean unsynchronized = false;
  private boolean noLog = false;
  private boolean reportTournament = false;
  private boolean fullScreen = false;
  private int fullScreenWidth;
  private int fullScreenHeight;
  private BufferStrategy bufferStrategy = null;
  private Font statusFont;
  private Graphics currentGraphics;
  
  private Color statusColor = Color.green.brighter();
  private Color statusColorHome = Color.blue.brighter();
  private Color statusColorVisitor = Color.red.brighter();
  
  private Server server;

  private boolean silent = true;
  private AudioClip 
    shotSound1      = null,
    shotSound2      = null,
    shotSound3      = null,
    rinkSound1      = null,
    rinkSound2      = null,
    rinkSound3      = null,
    tackleSound1    = null,
    tackleSound2    = null,
    postSound       = null,
    crowdSound      = null,
    collisionSound  = null,
    whistleSound    = null;

  private static final int NO_REASON                    = 0;
  private static final int REASON_HOME_GOALIE_KEEPS     = 1;
  private static final int REASON_VISITOR_GOALIE_KEEPS  = 2;
  private static final int REASON_HOME_SCORES           = 3;
  private static final int REASON_VISITOR_SCORES        = 4;
  private static final int REASON_HOME_OFFSIDE          = 5;
  private static final int REASON_VISITOR_OFFSIDE       = 6;
  private static final int REASON_HOME_ICING            = 7;
  private static final int REASON_VISITOR_ICING         = 8;
  private static final int REASON_HOME_INTERFERE        = 9;
  private static final int REASON_VISITOR_INTERFERE     = 10;
  private static final int REASON_HOME_FOUL             = 11;
  private static final int REASON_VISITOR_FOUL          = 12;
  private static final int REASON_UNPLAYABLE            = 13;
  
  private boolean gameOver = false;
  private boolean gameStopped = true;
  private boolean gameReady = false;
  private int currentFaceoffSpot;
  private int faceoffDelay = 0; 
  private int refereeReason = 0;
  private int currentFaceoffDelay = 0; 

  private boolean homeInOffside = false;
  private boolean visitorInOffside = false;
  private boolean homeInIcingThreat = false;
  private boolean visitorInIcingThreat = false;
  private boolean homeInIcing = false;
  private boolean visitorInIcing = false;
  private boolean homeInterferes = false;
  private boolean visitorInterferes = false;

  private static final Point2 HOME_GOAL_CENTER = 
    new Point2(0.0, -Rink.GOAL_LINE_OFFSET);
  private static final Point2 VISITOR_GOAL_CENTER = 
    new Point2(0.0, Rink.GOAL_LINE_OFFSET);

  private Player[] operationHistory = new Player[3];
  
  /**
   *  Creates a game which lasts <code>t</code> seconds,
   *  has an update delay of <code>s</code> seconds, and faceoff
   *  delay of <code>d</code> seconds.
   */
   
  Game(long t, int s, int d)
    {
    view = null;
    time = t;
    faceoffDelay = d;
    setUpdateFrequency(s);
    gameStopped = true;
    u = new Universe();
    rink = new Rink(u);
    puck = new Puck(PUCK_SAFE_POSITION);
    u.addBody(puck);

    try
      {
      u.setInelasticCollision(Class.forName("fi.utu.cs.hockey.Player"), 
                              Class.forName("fi.utu.cs.hockey.Puck"));
      u.setInelasticCollision(Class.forName("fi.utu.cs.hockey.rink.GoalNet"), 
                              Class.forName("fi.utu.cs.hockey.Puck"));
      }
    catch (ClassNotFoundException e)
      {
      throw new RuntimeException("Error in collision initialization: " + e);
      }

    teamHome = new Team();
    teamVisitor = new Team();
    teamHome.setTeam(true);
    teamVisitor.setTeam(false);

    try
      {

      for (int i = 0; i < Team.PLAYERS; i++)
        {
        Player pl = teamHome.addPlayer(i);
        pl.addCollisionNotification(
          Class.forName("fi.utu.cs.hockey.Player"));
        pl.addCollisionNotification(
          Class.forName("fi.utu.cs.hockey.rink.SideBoard"));
        pl.addCollisionNotification(
          Class.forName("fi.utu.cs.hockey.rink.CornerBoard"));
        pl.addCollisionNotification(
          Class.forName("fi.utu.cs.hockey.rink.EndBoard"));
        u.addBody(pl);
        }

      for (int i = 0; i < Team.PLAYERS; i++)
        {
        Player pl = teamVisitor.addPlayer(i);
        pl.addCollisionNotification(
          Class.forName("fi.utu.cs.hockey.Player"));
        pl.addCollisionNotification(
          Class.forName("fi.utu.cs.hockey.rink.SideBoard"));
        pl.addCollisionNotification(
          Class.forName("fi.utu.cs.hockey.rink.CornerBoard"));
        pl.addCollisionNotification(
          Class.forName("fi.utu.cs.hockey.rink.EndBoard"));
        u.addBody(pl);
        }

      puck.addCollisionNotification(
        Class.forName("fi.utu.cs.hockey.Player"));
      puck.addCollisionNotification(
        Class.forName("fi.utu.cs.hockey.rink.ScoreLine"));
      puck.addCollisionNotification(
        Class.forName("fi.utu.cs.hockey.rink.SideBoard"));
      puck.addCollisionNotification(
        Class.forName("fi.utu.cs.hockey.rink.CornerBoard"));
      puck.addCollisionNotification(
        Class.forName("fi.utu.cs.hockey.rink.EndBoard"));
      puck.addCollisionNotification(
        Class.forName("fi.utu.cs.hockey.rink.GoalPost"));
      puck.addCollisionNotification(
        Class.forName("fi.utu.cs.hockey.rink.Goal"));
      puck.addCollisionNotification(
        Class.forName("fi.utu.cs.hockey.rink.GoalCrease"));
      }
    catch (ClassNotFoundException e)
      {
      throw new RuntimeException("Error in collision initialization: " + e);
      }

    u.addObserver(new Collider());

    startGame();
    }

  
  /**
   *  Prepares a view to the game with the zoom factor <code>z</code>
   *  and rotates the view 90 degrees if <code>r == true</code>.
   */
  
  void prepareView(double z, boolean r)
    {
    setZoom(z);
    LinkedList list = new LinkedList(u.boundaries());
    list.addAll(u.bodies());
    view = new ShapeView(list, zoom, false);
    if (r) view.setRotated();
    view.setVisibleRegion(new Rectangle2D.Double(-15.0, -30.0, 30.0, 60.0));
    view.setBackground(Color.white);
    
    view.addFocusListener(new FocusAdapter() {
      public void focusGained(FocusEvent e)
        {
        if (!fullScreen)
          {
          view.repaint();
          view.revalidate();
          }
        }
      });
    }


  /** Sets the zoom factor to <code>z</code>. */
  private void setZoom(double z)
    {
    zoom = z;
    }


  /** Returns the zoom factor. */
  double getZoom()
    {
    return zoom;
    }


  /**
   *  Returns the view associated with the game, or <code>null</code>
   *  if it has not been prepared.
   *
   *  @see #prepareView(double,boolean)
   */
  ShapeView getView()
    {
    return view;
    }


  /**
   *  Sets a full screen mode to <code>width</code>  
   *  <code>height</code> using buffer <code>bs</code> 
   *  for page flipping, and text colors <code>hc</code>,
   *  <code>vc</code> and <code>rc</code>. 
   */
  void setFullScreen(int width, int height, BufferStrategy bs, 
                    String fontName, int fontSize, 
                    Color hc, Color vc, Color rc)
    {
    view.setIgnoreRepaint(true);
    fullScreen = true;
    bufferStrategy = bs;
    fullScreenWidth = width;
    fullScreenHeight = height;  
  	statusColor = rc;
  	statusColorHome = hc;
  	statusColorVisitor = vc;    

    Graphics g = bufferStrategy.getDrawGraphics();    
    statusFont = g.getFont();
    if (fontName != null)
      {
      statusFont = new Font(fontName, Font.PLAIN, fontSize); 
      }
    else if (fontSize > 0)
      {     
      statusFont = statusFont.deriveFont((float)fontSize);
      }    
    }
  
  
  /** Loads the sound files. */
  void loadSounds()
    {
    silent = false;
    try
      {
      String base = "file:" + System.getProperty("user.dir") + "/sounds/";
      shotSound1      = Applet.newAudioClip(new URL(base + "AIH_ShotHeavy.wav"));
      shotSound2      = Applet.newAudioClip(new URL(base + "AIH_ShotMedium.wav"));
      shotSound3      = Applet.newAudioClip(new URL(base + "AIH_ShotLight.wav"));
      rinkSound1      = Applet.newAudioClip(new URL(base + "AIH_RinkHeavy.wav"));
      rinkSound2      = Applet.newAudioClip(new URL(base + "AIH_RinkMedium.wav"));
      rinkSound3      = Applet.newAudioClip(new URL(base + "AIH_RinkLight.wav"));
      tackleSound1    = Applet.newAudioClip(new URL(base + "AIH_TackleHeavy.wav"));
      tackleSound2    = Applet.newAudioClip(new URL(base + "AIH_TackleHeavy.wav"));
      postSound       = Applet.newAudioClip(new URL(base + "AIH_Post.wav"));
      crowdSound      = Applet.newAudioClip(new URL(base + "AIH_Crowd.wav"));
      collisionSound  = Applet.newAudioClip(new URL(base + "AIH_Collision.wav"));
      whistleSound    = Applet.newAudioClip(new URL(base + "AIH_Whistle.wav"));
      }
    catch (MalformedURLException e)
      {
      System.err.println("Couldn't load sound file: " + e);
      }
    }


  /** Sets the update delay to <code>ms</code> milliseconds. */
  private void setUpdateFrequency(int ms)
    {
    timeSliceMs = ms;
    timeSliceS = ms / 1000.0;
    }

  /** Returns the update delay in milliseconds. */
  int getUpdateFrequency()
    {
    return timeSliceMs;
    }

  /** 
   *  Sets the game to be unsynchronized. The game time and the real-time
   *  are not synchronized (i.e., the game runs as fast as it can).
   */
  void setUnsynchronized()
    {
    unsynchronized = true;
    }

  /** Sets the game events not to be logged. */
  void setNoLog()
    {
    noLog = true;
    }

  /** Sets the log to be in tournament mode. */
  void setReportTournament()
    {
    reportTournament = true;
    }

  /** Advances the game state. Updates the view, if it exists. */
  private void advance()
    {
    u.advance(timeSliceS);
    if (view != null)
      {
      LinkedList list = new LinkedList(u.boundaries());
      list.addAll(u.bodies());

      view.setShapes(list);
      if (!fullScreen)
        {
        view.repaint();
        view.revalidate();
        }
      else
        {
        currentGraphics = bufferStrategy.getDrawGraphics();
        currentGraphics.setColor(Color.black);
        currentGraphics.fillRect(0, 0, fullScreenWidth, fullScreenHeight);
        view.render(currentGraphics);
        renderStatus(currentGraphics);
        currentGraphics.dispose();
        bufferStrategy.show();
        }
      }
    }



// ----------------------------------------------------
// ----------- METHODS FOR RUNNING THE GAME -----------
// ----------------------------------------------------

  /** Runs the game. */
  public void run()
    {
    updateStatus();
    
    long lastUpdateTime = time;
    
    while (!gameOver)
      {
      advance();
      if (gameReady)
        {
        if (gameStopped)
          {
          // Game is stopped, wait for the faceoff.
          currentFaceoffDelay -= timeSliceMs;
          if (currentFaceoffDelay <= 0)
            {
            // Waiting is over, do the faceoff.
            faceoff();
            updateStatusReferee();
            }
          }
        else
          {
          // Game is on, update the clock, uphold the rules, and
          // send updates to clients.

          time -= (long)timeSliceMs;

          // Is the game over?
          if (time <= 0L) 
            {
            stopGame();
            updateStatusReferee();
            }
          
          checkRules();
          
          if ((lastUpdateTime - time) >= 1000L)
            {
            // A second has past since the last update, update the clock.
            updateStatusTime();
            lastUpdateTime = time;
            }         
          }
        }
      server.sendGameState();
      // Sleep the rest of the loop 
      try
        {
        if (unsynchronized) 
          {
          sleep(FAST_TIME_SLICE_MS);
          }
        else sleep(timeSliceMs);
        }
      catch (InterruptedException e) { }
      }
    
    // Game over, send messages repeatedly 
    for (int i = 0; i < GAME_OVER_REPEATS; i++)
      {
      server.sendGameState();
      // Sleep a while
      try
        {
        yield();
        sleep(GAME_OVER_DELAY);
        }
      catch (InterruptedException e) { }
      }
    }



  /** Starts the game. */
  private void startGame()
    {
    gameReady = false;
    gameStopped = true;
    gameOver = false;
    currentFaceoffSpot = FaceOff.CENTER_SPOT;
    }


  /** Stops the game. After this call the game is over. */
  private void stopGame()
    {
    gameOver = true;
    addLogStop();
    }


  /** Stops the game and sets the next faceoff to the spot <code>s</code>. */
  void whistleFaceoff(int s)
    {
    whistleFaceoff(s, NO_REASON);
    }

  /** 
   *  Stops the game, sets the next faceoff to the spot <code>s</code>,
   *  and indicates that the reason is <code>r</code>.
   */
  private void whistleFaceoff(int s, int r)
    {
    if (!gameReady) addLogStart();
    if (whistleSound != null) whistleSound.play();
    else if (!silent) Toolkit.getDefaultToolkit().beep();
    gameReady = true;
    gameStopped = true;
    homeInOffside = false;
    visitorInOffside = false;
    homeInIcing = false;
    visitorInIcing = false;
    homeInIcingThreat = false;
    visitorInIcingThreat = false;
    homeInterferes = false;
    visitorInterferes = false;
    currentFaceoffSpot = s;
    currentFaceoffDelay = faceoffDelay;
    refereeReason = r;
    updateStatusReferee();
    }


  /** Realizes a faceoff. */
  private void faceoff()
    {
    gameStopped = false;
    refereeReason = NO_REASON;
    teamHome.faceoff(currentFaceoffSpot);
    teamVisitor.faceoff(currentFaceoffSpot);
    puck.faceoff(currentFaceoffSpot);
    clearOperationHistory();
    }


// -------------------------------------------------------
// ----------- METHODS FOR UPHOLDING THE RULES -----------
// -------------------------------------------------------


  /** Checks whether any rule is being broken. */
  private void checkRules()
    {
    checkOffside();
    checkIcing();
    checkUnplayability();
    }


  /** Check whether offside rules are being broken. */   
  private void checkOffside()
    {
    if (!gameStopped)
      {
      if (homeInOffside && 
          teamHome.forAllPlayersLessThan(Rink.BLUE_LINE_OFFSET))
        {
        homeInOffside = false;
        updateStatusReferee();
        }
      else if (!homeInOffside && 
          puck.getPosition().getY() < Rink.BLUE_LINE_OFFSET &&
          teamHome.existsPlayerGreaterThan(Rink.BLUE_LINE_OFFSET))
        {
        homeInOffside = true;
        updateStatusReferee();
        }

      if (visitorInOffside && 
          teamVisitor.forAllPlayersGreaterThan(-Rink.BLUE_LINE_OFFSET))
        {
        visitorInOffside = false;
        updateStatusReferee();
        }
      else if (!visitorInOffside &&
          puck.getPosition().getY() > -Rink.BLUE_LINE_OFFSET &&
          teamVisitor.existsPlayerLessThan(-Rink.BLUE_LINE_OFFSET))
        {
        visitorInOffside = true;
        updateStatusReferee();
        }
      }
    }


  /** Check whether player <code>p</code> is in offside. */
  private boolean checkPlayerOffside(Player p)
    {
    if (!gameStopped)
      {
      if (homeInOffside && p.isHome() 
          && p.getPosition().getY() > Rink.BLUE_LINE_OFFSET)
        {
        int s = (p.getPosition().getX() < 0.0 ?
                 FaceOff.VISITOR_NEUTRAL_ZONE_LEFT :
                 FaceOff.VISITOR_NEUTRAL_ZONE_RIGHT);
        whistleFaceoff(s, REASON_HOME_OFFSIDE);     
        return true;
        }
      if (visitorInOffside && !p.isHome() 
          && p.getPosition().getY() < -Rink.BLUE_LINE_OFFSET)
        {
        int s = (p.getPosition().getX() < 0.0 ?
                 FaceOff.HOME_NEUTRAL_ZONE_LEFT :
                 FaceOff.HOME_NEUTRAL_ZONE_RIGHT);
        whistleFaceoff(s, REASON_VISITOR_OFFSIDE);
        return true;
        }
      }
    return false;
    }


  /** Check whether icing rules are being broken. */
  private void checkIcing()
    {
    if (!gameStopped)
      {
      if (homeInIcingThreat 
        && puck.getPosition().getY() > Rink.BLUE_LINE_OFFSET)
        {
        homeInIcing = true;
        updateStatusReferee();
        }
      if (homeInIcing 
        && puck.getPosition().getY() > Rink.GOAL_LINE_OFFSET)
        {
        int s = (puck.getPosition().getX() < 0.0 ?
                 FaceOff.HOME_ENDZONE_LEFT :
                 FaceOff.HOME_ENDZONE_RIGHT);
        whistleFaceoff(s, REASON_HOME_ICING);
        }

      if (visitorInIcingThreat 
        && puck.getPosition().getY() < -Rink.BLUE_LINE_OFFSET)
        {
        visitorInIcing = true;
        updateStatusReferee();
        }
      if (visitorInIcing 
        && puck.getPosition().getY() < -Rink.GOAL_LINE_OFFSET)
        {
        int s = (puck.getPosition().getX() < 0.0 ?
                 FaceOff.VISITOR_ENDZONE_LEFT :
                 FaceOff.VISITOR_ENDZONE_RIGHT);
        whistleFaceoff(s, REASON_VISITOR_ICING);
        }
      }
    }


  /** Check whether player <code>p</code> is in icing. */
  private void checkPlayerIcing(Player p)
    {
    if (!gameStopped)
      {
      if (p.isHome())
        {
        clearIcing(false);
        if (p.getPosition().getY() < 0.0)
          homeInIcingThreat = true;
        else
          clearIcing(true);
        }
      else if (!p.isHome())
        {
        clearIcing(true);
        if (p.getPosition().getY() > 0.0)
          visitorInIcingThreat = true;
        else
          clearIcing(false);
        } 
      }
    }

  
  /** Clears icing for the team <code>h</code>. */
  private void clearIcing(boolean h)
    {
    if (h) 
      {
      homeInIcing = false;
      homeInIcingThreat = false;
      updateStatusReferee();
      }
    else 
      {
      visitorInIcing = false;
      visitorInIcingThreat = false;
      updateStatusReferee();
      }
    }


  /** Check whether a goalie is being interfered. */
  private void checkPlayerInterference(Player p1, Player p2)
    {
    if (!gameStopped)
      {
      if (p1.isHome() == p2.isHome()) return;
      if (p1.getPlayPosition() == p2.getPlayPosition()) return;
      if (p2.getPlayPosition() == Position.GOALKEEPER)
        {
        Player tmp = p1;
        p1 = p2;
        p2 = tmp;
        }
      if (p1.getPlayPosition() == Position.GOALKEEPER)
        {
        if (p1.isHome())
          {
          Vector2 v = new Vector2(p1.getPosition(), HOME_GOAL_CENTER);
          if (v.length() <= Rink.GOAL_CREASE_RADIUS)
            {
            visitorInterferes = true;
            updateStatusReferee();
            }
          }
        else if (!p1.isHome())
          {
          Vector2 v = new Vector2(p1.getPosition(), VISITOR_GOAL_CENTER);
          if (v.length() <= Rink.GOAL_CREASE_RADIUS)
            {
            homeInterferes = true;
            updateStatusReferee();
            }
          }
        }
      }
    }


  /** 
   *  Check whether there is a pending interference whistle for
   *  the team of player <code>p</code>.
   */
  private boolean checkPendingPlayerInterference(Player p)
    {
    if (!gameStopped)
      {
      if (homeInterferes && p.isHome())
        {
        int s = (p.getPosition().getX() < 0.0 ?
                 FaceOff.HOME_ENDZONE_LEFT :
                 FaceOff.HOME_ENDZONE_RIGHT);
        whistleFaceoff(s, REASON_HOME_INTERFERE);     
        return true;
        }
      if (visitorInterferes && !p.isHome())
        {
        int s = (p.getPosition().getX() < 0.0 ?
                 FaceOff.VISITOR_ENDZONE_LEFT :
                 FaceOff.VISITOR_ENDZONE_RIGHT);
        whistleFaceoff(s, REASON_VISITOR_INTERFERE);      
        return true;
        }
      }
    return false;
    }

  /** Check whether the puck has become unplayable. */
  private void checkUnplayability()
    {
    if (!gameStopped)
      {
      double x = puck.getPosition().getX();
      double y = puck.getPosition().getY();
      if (x < -Rink.WIDTH / 2.0 | x > Rink.WIDTH / 2.0
        | y < -Rink.LENGTH / 2.0 | y > Rink.LENGTH / 2.0)
        whistleFaceoff(FaceOff.CENTER_SPOT, REASON_UNPLAYABLE);
      }
    }
    
  private void updateStatus()
    {
    updateStatusTime();
    updateStatusScore();
    updateStatusRoster();
    updateStatusReferee();    
    }



// -----------------------------------------------------------
// ----------- METHODS FOR QUERYING THE GAME STATE -----------
// -----------------------------------------------------------


  /** Returns the home team. */
  public Team getHome()
    {
    return teamHome;
    }
  
  /** Returns the visiting team. */
  public Team getVisitor()
    {
    return teamVisitor;
    }

  /** Returns the puck. */
  public Puck getPuck()
    {
    return puck;
    }

  /** Returns the time left in the game in milliseconds. */
  public long getTime()
    {
    return time;
    }
    
  /** 
   *  Returns the score for the home team if <code>home == true</code>,
   *  otherwise the score for the visiting team.
   */
  public int getScore(boolean home)
    {
    return (home ? scoreHome : scoreVisitor);
    }

  /** Sets the server to <code>s</code>. */
  void setServer(Server s)
    {
    server = s;
    }
    
  
  /** Signal that teams have been updated. */
  public void teamUpdated()
    {
    updateStatusRoster();
    }
  

  /** Returns <code>true</code>, if the game is stopped. */
  public boolean isGameStopped()
    {
    return gameStopped;
    }


  /** Returns <code>true</code>, if the game is over. */
  public boolean isGameOver()
    {
    return gameOver;
    }

  /** 
   *  Returns the faceoff spot. 
   *
   *  @see FaceOff
   */

  public int getFaceoffSpot()
    {
    return currentFaceoffSpot;
    }

  /** 
   *  Returns <code>true</code>, if <code>h == true</code> and
   *  the home team has scored, or if <code>h == false</code>
   *  and the visiting team has scored.
   */
  public boolean isGoal(boolean h)
    {
    return (h ? (refereeReason == REASON_HOME_SCORES) :
                (refereeReason == REASON_VISITOR_SCORES));
    }

  /** 
   *  Returns <code>true</code>, if <code>h == true</code> and
   *  the home team causes an offside, or if <code>h == false</code>
   *  and the visiting team causes an offside.
   */
  public boolean isOffside(boolean h)
    {
    return (h ? 
      (refereeReason == REASON_HOME_OFFSIDE | homeInOffside) :
      (refereeReason == REASON_VISITOR_OFFSIDE | visitorInOffside));
    }

  /** 
   *  Returns <code>true</code>, if <code>h == true</code> and
   *  the home team causes an icing, or if <code>h == false</code>
   *  and the visiting team causes an icing.
   */
  public boolean isIcing(boolean h)
    {
    return (h ? 
      (refereeReason == REASON_HOME_ICING | homeInIcing) :
      (refereeReason == REASON_VISITOR_ICING | visitorInIcing));
    }

  /** 
   *  Returns <code>true</code>, if <code>h == true</code> and
   *  the home team is interfering the goalie, or if <code>h == false</code>
   *  and the visiting team is interfering the goalie.
   */
  public boolean isInterferingGoalie(boolean h)
    {
    return (h ? 
      (refereeReason == REASON_HOME_INTERFERE | homeInterferes) :
      (refereeReason == REASON_VISITOR_INTERFERE | visitorInterferes));
    }

  /** 
   *  Returns <code>true</code>, if <code>h == true</code> and
   *  the home team goalie keeps the puck, or if <code>h == false</code>
   *  and the visiting team goalie keeps the puck.
   */
  public boolean isGoalieKeepingPuck(boolean h)
    {
    return (h ? (refereeReason == REASON_HOME_GOALIE_KEEPS) :
                (refereeReason == REASON_VISITOR_GOALIE_KEEPS));
    }

  /** 
   *  Returns <code>true</code>, if <code>h == true</code> and
   *  the home team commits a foul, or if <code>h == false</code>
   *  and the visiting team commits a foul.
   */
  public boolean isFoul(boolean h)
    {
    return (h ? (refereeReason == REASON_HOME_FOUL) :
                (refereeReason == REASON_VISITOR_FOUL));
    }
  
  /** 
   *  Returns <code>true</code>, if the puck has become
   *  unplayable.
   */
  public boolean isPuckUnplayable()
    {
    return (refereeReason == REASON_UNPLAYABLE);
    }



// ----------------------------------------------------------
// ----------- METHODS FOR GAME STATUS COMPONENTS -----------
// ----------------------------------------------------------
  

  /**
   *  Renders the status in full screen mode.
   */
  private void renderStatus(Graphics g)
    {
    Graphics2D g2 = (Graphics2D)g;
    g2.setTransform(new AffineTransform());
    g2.setClip(0, 0, fullScreenWidth, fullScreenHeight);

    g.setFont(statusFont);
    renderLine(g, statusColor, getTimeMMSS(), 4, 0);
    renderLine(g, statusColor, "" + scoreHome + "-" + scoreVisitor, 3, 0);
    if (teamHome.getName() != null)
      renderLine(g, statusColorHome, teamHome.getName(), 3, 1);
    if (teamVisitor.getName() != null)
      renderLine(g, statusColorVisitor, teamVisitor.getName(), 3, 2);

    if (gameOver) 
      renderLine(g, statusColor, "Game over.", 2, 0);
    else if (gameStopped) 
      {
      Color c = statusColor;
      String rt = null;
      switch (refereeReason)
        {
        case REASON_HOME_GOALIE_KEEPS:
          c = statusColorHome;
          rt = "Goalie keeps the puck.";
          break;
        case REASON_VISITOR_GOALIE_KEEPS:
          c = statusColorVisitor;
          rt = "Goalie keeps the puck.";
          break;
        case REASON_HOME_SCORES:
          c = statusColorHome;
          rt = "Goal!";
          break;
        case REASON_VISITOR_SCORES:
          c = statusColorVisitor;
          rt = "Goal!";
          break;            
        case REASON_HOME_OFFSIDE:
          c = statusColorHome;
          rt = "Offside.";
          break;        
        case REASON_VISITOR_OFFSIDE:
          c = statusColorVisitor;
          rt = "Offside.";
          break;
        case REASON_HOME_ICING:
          c = statusColorHome;
          rt = "Icing.";
          break;
        case REASON_VISITOR_ICING:
          c = statusColorVisitor;
          rt = "Icing.";
          break;
        case REASON_HOME_INTERFERE:
          c = statusColorHome;
          rt = "Interfering the goalie.";
          break;
        case REASON_VISITOR_INTERFERE:
          c = statusColorVisitor;
          rt = "Interfering the goalie.";
          break;
        case REASON_HOME_FOUL:
          c = statusColorHome;
          rt = "Foul.";
          break;
        case REASON_VISITOR_FOUL:
          c = statusColorVisitor;
          rt = "Foul.";
          break;
        case REASON_UNPLAYABLE:
          rt = "Puck is unplayable.";
          break;
        }
      if (rt != null) renderLine(g, c, rt, 2, 0);
      }
    else
      {     
      String result = "";
      if (homeInOffside)
        result = result + " offside ";
      if (homeInIcing)
        result = result + " icing ";
      if (homeInterferes)
        result = result + " interference ";
      renderLine(g, statusColorHome, result, 2, 1);

      result = "";
      if (visitorInOffside) 
        result = result + " offside ";
      if (visitorInIcing)
        result = result + " icing ";
      if (visitorInterferes)
        result = result + " interference ";
      renderLine(g, statusColorVisitor, result, 2, 2);
      }

    if (!gameStopped)
      {
      Player p = lastOperatingPlayer();
      String result = "";
      Color c = statusColor;
      if (p != null)
        {
        result = "#" + p.getNumber() + " " + p.getName();
        c = (p.isHome() ? statusColorHome : statusColorVisitor);
        }
      renderLine(g, c, result, 1, 0);
      }
    else if (refereeReason == REASON_HOME_SCORES)
      {
      renderLine(g, statusColorHome, getScorer(teamHome), 1, 0);
      }
    else if (refereeReason == REASON_VISITOR_SCORES)
      {
      renderLine(g, statusColorVisitor, getScorer(teamVisitor), 1, 0);
      }
    }


  /** Renders one line of text in full screen mode. */
  private void renderLine(Graphics g, Color c, String s, int l, int p)
    {
    g.setColor(c);
    FontMetrics fm = g.getFontMetrics();
    int w = fm.stringWidth(s);
    int h = fm.getHeight();
    int x = (fullScreenWidth - w) / 2;
    if (p == 1)
      {
      x = ((fullScreenWidth / 2) - w) / 2;
      }
    else if (p == 2)
      {
      x = (((fullScreenWidth / 2) - w) / 2) + (fullScreenWidth / 2);
      }
    g.drawString(s, x, fullScreenHeight - (l * h));   
    }
  
  /**
   *  Sets the graphical component for displaying the game status.
   *
   *  @param t Time 
   *  @param s Score
   *  @param r Roster
   *  @param f Referee
   */
  void setGameStatus(JLabel t, JLabel s, JLabel r, JLabel f)
    {
    gameStatusTime = t;
    gameStatusScore = s;
    gameStatusRoster = r;
    gameStatusReferee = f;
    }

  /** Updates the time component. */
  private void updateStatusTime()
    {
    if (gameStatusTime != null)
      {
      int min = (int)time / 60000;
      int sec = ((int)time / 1000) - (min * 60);
      gameStatusTime.setText("<html><h2>" + getTimeMMSS() + "</h2></html>");
      }
    }

  /** Returns the remaining time in format MM:SS. */
  private String getTimeMMSS()
    {
	// Can't avoid this ugly hack...
	long correctedTime = (fullScreen ? time * 2 : time); 

    int min = (int)correctedTime / 60000;
    int sec = ((int)correctedTime / 1000) - (min * 60);
    return "" + min + ":" + (sec < 10 ?  "0" : "") + sec;
    }

  /** Updates the score component. */
  private void updateStatusScore()
    {
    if (gameStatusScore != null)
      {
      gameStatusScore.setText("<html><h2>" + scoreHome + "-" 
                              + scoreVisitor + "</h2></html>");
      
      }
    }

  /** Updates the roster component. */
  private void updateStatusRoster()
    {
    if (gameStatusRoster != null)
      {
      gameStatusRoster.setText("<html>" + roster(teamHome) 
                              + roster(teamVisitor) + "</html>");
      }
    }

  /** Updates the referee component. */
  private void updateStatusReferee()
    {
    if (gameStatusReferee != null)
      {
      String result = "<html><center>";
      if (gameOver) result = result + "<h1>Game over</h1>";
      else if (!gameReady) result = result + "<h1>Game not started</h1>";
      else if (gameStopped) 
        {
        result = result + "<h1>Game stopped</h1>";
        
        String rt = null;
        switch (refereeReason)
          {
          case REASON_HOME_GOALIE_KEEPS:
          case REASON_VISITOR_GOALIE_KEEPS:
            rt = "goalie keeps the puck";
            break;
          case REASON_HOME_SCORES:
            rt = "home team scores";
            break;
          case REASON_VISITOR_SCORES:
            rt = "visiting team scores";
            break;            
          case REASON_HOME_OFFSIDE:
          case REASON_VISITOR_OFFSIDE:
            rt = "offside";
            break;
          case REASON_HOME_ICING:
          case REASON_VISITOR_ICING:
            rt = "icing";
            break;
          case REASON_HOME_INTERFERE:
          case REASON_VISITOR_INTERFERE:
            rt = "interfering the goalie";
            break;
          case REASON_HOME_FOUL:
            rt = "home team's foul";
            break;
          case REASON_VISITOR_FOUL:
            rt = "visiting team's foul";
            break;
          case REASON_UNPLAYABLE:
            rt = "puck is unplayable";
            break;
          }
        if (rt != null) result = result + "<h2>Reason: " + rt + "</h2>";
        
        String ft = "Faceoff: ";
        switch (currentFaceoffSpot)
          {
          case FaceOff.CENTER_SPOT: 
            ft = ft + "center circle";
            break;
          case FaceOff.HOME_ENDZONE_LEFT: 
          case FaceOff.HOME_ENDZONE_RIGHT: 
            ft = ft + "home team's endzone";
            break;
          case FaceOff.HOME_NEUTRAL_ZONE_LEFT: 
          case FaceOff.HOME_NEUTRAL_ZONE_RIGHT: 
            ft = ft + "home team's neutral zone";
            break;
          case FaceOff.VISITOR_ENDZONE_LEFT: 
          case FaceOff.VISITOR_ENDZONE_RIGHT: 
            ft = ft + "visiting team's endzone";
            break;
          case FaceOff.VISITOR_NEUTRAL_ZONE_LEFT: 
          case FaceOff.VISITOR_NEUTRAL_ZONE_RIGHT: 
            ft = ft + "visiting team's neutral zone";
            break;
          }
        result = result + "<h2>" + ft + "</h2>";
        }
      else
        {
        if (homeInOffside)
          result = result + "<h3>Referee: offside (home)</h3>";
        if (visitorInOffside) 
          result = result + "<h3>Referee: offside (visitor)</h3>";
        if (homeInIcing)
          result = result + "<h3>Referee: icing (home)</h3>";
        if (visitorInIcing)
          result = result + "<h3>Referee: icing (visitor)</h3>";
        if (homeInterferes)
          result = result + "<h3>Referee: interference (home)</h3>";
        if (visitorInterferes)
          result = result + "<h3>Referee: interference (visitor)</h3>";
        }
      gameStatusReferee.setText(result + "</center></html>");
      }
    }


  /** Returns the roster for team <code>t</code>. */
  private String roster(Team t)
    {
    String result = "<h2>" + t.getName() + "</h2><table>";
    for (int i = 0; i < 6; i++)
      {
      Player p = t.getPlayer(i);
      result = result + "<tr><td valign=top>" + POSITION_SHORTCUT[i] 
              + "</td><td valign=top>#" + p.getNumber() 
              + "</td><td valign=top>" + p.getName() + "</td></tr>";
      }
    result = result + "</table>";
    return result;
    }



// -----------------------------------------------------
// ----------- METHODS FOR PLAYER OPERATIONS -----------
// -----------------------------------------------------


  /** 
   *  Lets the player <code>p</code> shoot the puck with the proportional
   *  power <code>s</code>, if possible.
   */ 
  public void shoot(Player p, double s)
    {
    if (!gameStopped && p.canReach(puck))
      {
      if (!checkPlayerOffside(p))
        {
        if (!checkPendingPlayerInterference(p))
          {
          checkPlayerIcing(p);
          Vector2 v = p.getOrientation().unitVector();          
          v.multiply(s * Player.MAX_SHOT_VELOCITY);
          // Play sound
          if (s > 0.66)
            {
            if (shotSound1 != null) shotSound1.play();
            }
          else if (s > 0.33)
            {
            if (shotSound2 != null) shotSound2.play();
            }
          else 
            {
            if (shotSound3 != null) shotSound3.play();
            }
          // Add noise to the shot
          v.rotate(2.0 * (random.nextDouble() - 0.5) * s * MAX_PUCK_NOISE);
          puck.setVelocity(v);
          puck.setAcceleration(0.0,0.0);
          addOperationHistory(p);
          }
        }
      }
    }


  /** 
   *  Lets the goalkeeper <code>p</code> keep the puck, if possible.
   */   
  public void keepPuck(Player p)
    {
    if (!gameStopped && p.getPlayPosition() == Position.GOALKEEPER &&
        p.canReach(puck))
      {
      if (!checkPendingPlayerInterference(p))
        {
        double y = p.getPosition().getY();
        if ((p.isHome() & 
            y >= -Rink.GOAL_LINE_OFFSET & y <= -Rink.FACE_OFF_OFFSET)
         || (!p.isHome() &
            y <= Rink.GOAL_LINE_OFFSET & y >= Rink.FACE_OFF_OFFSET))
          {
          int s = FaceOff.CENTER_SPOT;
          int r = NO_REASON;
          if (p.isHome()) 
            {
            s = ((puck.getPosition().getX() < 0.0) ?
                  FaceOff.HOME_ENDZONE_LEFT :
                  FaceOff.HOME_ENDZONE_RIGHT);
            r = REASON_HOME_GOALIE_KEEPS;
            }
          else 
            {
            s = ((puck.getPosition().getX() < 0.0) ?
                  FaceOff.VISITOR_ENDZONE_LEFT :
                  FaceOff.VISITOR_ENDZONE_RIGHT);
            r = REASON_VISITOR_GOALIE_KEEPS;
            }
          puck.setPosition(PUCK_SAFE_POSITION);
          whistleFaceoff(s, r);
          }
        }
      }
    }


// -----------------------------------------
// ----------- OPERATION HISTORY -----------
// -----------------------------------------

  /** Adds player <code>p</code> to the operation history. */
  private void addOperationHistory(Player p)
    {
    if (operationHistory[0] != null)
      {
      if (operationHistory[0] == p) return;
      if (operationHistory[0].isHome() != p.isHome())
        {
        clearOperationHistory();        
        }
      else
        {
        for (int i = operationHistory.length - 1; i >= 1; i--)
          {
          if (operationHistory[i - 1] == p) 
            operationHistory[i] = null;
          else 
            operationHistory[i] = operationHistory[i - 1];
          }
        }
      }
    operationHistory[0] = p;
    }


  /** Clears the operation history. */
  private void clearOperationHistory()
    {
    for (int i = 0; i < operationHistory.length; i++)
      operationHistory[i] = null;   
    }


  /** Returns the players in operation history. */
  private Player[] retrieveOperationHistory()
    {
    int last = 0;
    while (last < operationHistory.length && 
           operationHistory[last] != null) last++;
    if (last == 0) return null;
    Player[] result = new Player[last];
    for (int i = 0; i < last; i++)
      result[i] = operationHistory[i];
    return result;
    }

  /** Returns the last player that have touched the puck. */
  private Player lastOperatingPlayer()
    {
    return operationHistory[0];
    }


// --------------------------------------------------
// ----------- METHODS FOR KEEPING A LOG  -----------
// --------------------------------------------------

  private void addLogStart()
    {
    if (noLog | reportTournament) return;
    System.out.println(teamHome.getName() + " vs. " + teamVisitor.getName());
    System.out.println();
    }

  private void addLogStop()
    {
    if (noLog) return;
    if (!reportTournament)
      {
      System.out.println();
      System.out.println("Final: " + scoreHome + "-" + scoreVisitor);
      }
    else
      {
      System.out.println(teamHome.getName() + ";" +
												 teamVisitor.getName() + ";" +
												 scoreHome + ";" + scoreVisitor);
      }
    }

  private void addLogScore(Team t)
    {
    if (noLog) return;
    if (!reportTournament)
    	{
    	String result = "" + scoreHome + "-" + scoreVisitor + " (" 
    	                + getTimeMMSS() + ") " + getScorer(t);
    	System.out.println(result);
    	}
    else
    	{
    	String s = getScorerTournament(t);
    	if (s != null) System.out.println(s);
    	}
    }
    

  private String getScorer(Team t)    
    {
    String result = "";
    Player[] history = retrieveOperationHistory();
    if (history != null && history.length > 0 && 
        t.isHome() == history[0].isHome())
      {
      Player p = history[0];
      result = result + "#" + p.getNumber() + " " + p.getName();
      for (int i = 1; i < history.length; i++)
        {
        if (i == 1) result = result + " (";
        p = history[i];
        result = result + "#" + p.getNumber() + " " + p.getName();
        if (i != (history.length - 1)) result = result + ", ";
        else result = result + ")";
        }
      }
    else 
      {
      result = result + "own goal";
      }
    return result;
    }



  private String getScorerTournament(Team t)    
    {
    String result = t.getName() + ";";
    Player[] history = retrieveOperationHistory();
    if (history != null && history.length > 0 && 
        t.isHome() == history[0].isHome())
      {
      Player p = history[0];
      result = result + "#" + p.getNumber() + " " + p.getName() + ";";
      
      if (history.length > 1)
      	{
        p = history[1];
      	result = result + "#" + p.getNumber() + " " + p.getName() + ";";
      	}
      else result = result + ";";

      if (history.length > 2)
      	{
        p = history[2];
      	result = result + "#" + p.getNumber() + " " + p.getName();
      	}
      }
    else 
      {
      result = null;
      }
    return result;
    }


// -------------------------------------------
// ----------- COLLISION DETECTION -----------
// -------------------------------------------

  /**
   *  Checks an observed collision and reacts accordingly. 
   */
  private class Collider implements Observer
    {
    public void update(Observable o, Object m)
      {
      Universe.CollisionNotification c = (Universe.CollisionNotification)m;
      if (c.getParticle() instanceof SideBoard 
       || c.getParticle() instanceof CornerBoard
       || c.getParticle() instanceof EndBoard)
        {
        // something to board
        if (c.getBody() instanceof Puck)
          {
          // puck to board
          double force = puck.getVelocity().length();
          force /= Player.MAX_SHOT_VELOCITY;
          if (force > 0.50)
            {
            if (rinkSound1 != null) rinkSound1.play();
            }
          else if (force > 0.30)
            {
            if (rinkSound2 != null) rinkSound2.play();
            }
          else if (force > 0.10)
            {
            if (rinkSound3 != null) rinkSound3.play();
            }
          }
        if (c.getBody() instanceof Player)
          {
          // player to board
          if (random.nextBoolean())
            {
            if (tackleSound1 != null) tackleSound1.play();
            }
          else
            {
            if (tackleSound2 != null) tackleSound2.play();
            }
          }
        }
      else if (c.getBody() instanceof Player 
            && c.getParticle() instanceof Player)
        {
        // player to player
        Player p1 = (Player)c.getBody();
        Player p2 = (Player)c.getParticle();
        checkPlayerInterference(p1, p2);
        if (p1.isHome() != p2.isHome() & 
            collisionSound != null) collisionSound.play();
        }
      else if (c.getBody() instanceof Puck 
            && (c.getParticle() instanceof GoalPost
              | c.getParticle() instanceof Goal))
        {
        // puck to post
        if (postSound != null) postSound.play();
        }
      else if (c.getBody() instanceof Puck 
            && (c.getParticle() instanceof GoalCrease))
        {
        // puck to goal crease
        if (homeInIcing && puck.getPosition().getY() > 0.0)
          clearIcing(true);
        if (visitorInIcing && puck.getPosition().getY() < 0.0)
          clearIcing(false);
        }
      else if (c.getBody() instanceof Puck 
            && (c.getParticle() instanceof Player))
        {
        // puck to player
        Player p = (Player)c.getParticle();
        checkPlayerOffside(p);
        checkPlayerIcing(p);
        checkPendingPlayerInterference(p);
        }       
      else if (c.getBody() instanceof Player 
            && (c.getParticle() instanceof Puck))
        {
        // player to puck
        Player p = (Player)c.getBody();
        checkPlayerOffside(p);
        checkPlayerIcing(p);
        checkPendingPlayerInterference(p);
        }       
      else if (!gameStopped
            && c.getBody() instanceof Puck 
            && c.getParticle() instanceof ScoreLine)
        {
        // puck to scoreline
        clearIcing(true);
        clearIcing(false);
        int r = NO_REASON;
        int s = FaceOff.CENTER_SPOT;
        if (c.getParticle() == rink.scoreLines[0]) 
          {
          // puck to home scoreline
          if (!visitorInOffside & !visitorInterferes)
            {
            // visitor scores
            scoreVisitor++;
            r = REASON_VISITOR_SCORES;
            updateStatusScore();
            if (crowdSound != null) crowdSound.play();
            addLogScore(teamVisitor);
            }
          else if (visitorInterferes)
            {
            // pending interference -> no score
            s = (puck.getPosition().getX() < 0.0 ?
                     FaceOff.VISITOR_ENDZONE_LEFT :
                     FaceOff.VISITOR_ENDZONE_RIGHT);
            r = REASON_VISITOR_INTERFERE;     
            }
          else 
            {
            // offside -> no score
            s = (puck.getPosition().getX() < 0.0 ?
                     FaceOff.HOME_NEUTRAL_ZONE_LEFT :
                     FaceOff.HOME_NEUTRAL_ZONE_RIGHT);
            r = REASON_VISITOR_OFFSIDE;
            }       
          }
        if (c.getParticle() == rink.scoreLines[1]) 
          {
          // puck to visitor scoreline
          if (!homeInOffside & !homeInterferes)
            {
            // home scores
            scoreHome++;
            r = REASON_HOME_SCORES;
            updateStatusScore();
            if (crowdSound != null) crowdSound.play();
            addLogScore(teamHome);
            }
          else if (homeInterferes)
            {
            // pending interference -> no score
            s = (puck.getPosition().getX() < 0.0 ?
                     FaceOff.HOME_ENDZONE_LEFT :
                     FaceOff.HOME_ENDZONE_RIGHT);
            r = REASON_HOME_INTERFERE;      
            }
          else
            {
            // offside -> no score
            s = (puck.getPosition().getX() < 0.0 ?
                     FaceOff.VISITOR_NEUTRAL_ZONE_LEFT :
                     FaceOff.VISITOR_NEUTRAL_ZONE_RIGHT);
            r = REASON_HOME_OFFSIDE;
            }
          }
        whistleFaceoff(s, r);
        }
      }
    }

  }