/* 
 * Body.java 1.1 2002-06-11
 *
 *  2001-2002 Jouni Smed
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of the University of Turku nor the names of its
 *       contributors may be used to endorse or promote products derived 
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package fi.utu.cs.physics;

import java.util.Set;
import java.util.HashSet;

/**
 *  A moving physical body with mass and dimensionality. The outline of 
 *  the body is approximated with a bounding circle of a given radius, 
 *  which is used in detecting overlaps. The body can be owned by
 *  an instance of the class <code>Universe</code>, which may impose limits
 *  to the position. 
 * 
 *  @see Universe
 *  @version 1.1, 020611
 *  @author  Jouni Smed
 */

public abstract class Body extends Particle
  {    
  /** The mass of the body. */
  private double mass;
  /** The radius of the bounding circle. */
  private double radius;
  /** The current position of the body. */
  private Point2 position;
  /** The current velocity of the body. */
  private Vector2 velocity;
  /** The current acceleration of the body. */
  private Vector2 acceleration;

  /** A set comprising currently overlapping bodies. */
  private Set currentOverlaps = new HashSet();  
  /** A set comprising overlapping bodies from the previous update. */
  private Set previousOverlaps = new HashSet(); 

  
  /** 
   *  Constructs a body with the mass <code>m</code> and 
   *  the radius <code>r</code>. The body resides in the 
   *  origin and has a zero velocity and a zero acceleration.
   *  The body is corporeal and has no owner.  
   */
  protected Body(double m, double r)
    {
    this(m, r, Point2.ORIGIN, Vector2.ZERO, Vector2.ZERO);
    }

  /** 
   *  Constructs a body with the mass <code>m</code>,
   *  the radius <code>r</code> and the position <code>p</code>. 
   *  The body has a zero velocity and a zero acceleration.
   *  The body is corporeal and has no owner.  
   */
  protected Body(double m, double r, Point2 p)
    {
    this(m, r, p, Vector2.ZERO, Vector2.ZERO);
    }
  
  /** 
   *  Constructs a body with the mass <code>m</code>,
   *  the radius <code>r</code>, the position <code>p</code>
   *  and the velocity <code>v</code>.  The body has a 
   *  zero acceleration.   
   *  The body is corporeal and has no owner.  
   */
  protected Body(double m, double r, Point2 p, Vector2 v)
    {
    this(m, r, p, v, Vector2.ZERO);
    }

  /** 
   *  Constructs a body with the mass <code>m</code>,
   *  the radius <code>r</code>, the position <code>p</code>,
   *  the velocity <code>v</code> and the acceleration <code>a</code>.
   *  The body is corporeal and has no owner.  
   */
  protected Body(double m, double r, Point2 p, Vector2 v, Vector2 a)
    {
    super();
    mass         = m;
    radius       = r;
    position     = (Point2)p.clone();
    velocity     = (Vector2)v.clone();
    acceleration = (Vector2)a.clone();
    setOwner(null);
    }

  /** Returns the mass of the body. */
  public double getMass()
    {
    return mass;
    }

  /** Returns the radius of the bounding circle. */
  public double getRadius()
    {
    return radius;
    }

  /** Returns the position of the body. */
  public Point2 getPosition()
    {
    return (Point2)position.clone();
    }

  /** Returns the velocity of the body. */
  public Vector2 getVelocity()
    {
    return (Vector2)velocity.clone();
    }

  /** Returns the acceleration of the body. */
  public Vector2 getAcceleration()
    {
    return (Vector2)acceleration.clone();
    }

  /** Sets the mass of the body to <code>m</code>. */
  public void setMass(double m)
    {
    mass = m;
    }
  
  /** Sets the radius of the bounding circle to <code>r</code>. */
  public void setRadius(double r)
    {
    radius = r;
    }

  /** 
   *  Sets the position of the body. 
   *  @param x The x coordinate of the position.
   *  @param y The y coordinate of the position.
   *  @see #setPosition(Point2)
   */
  public void setPosition(double x, double y)
    {
    if (getOwner() != null) 
    	position.set(getOwner().transpose(new Point2(x, y)));
    else position.set(x, y);
    }

  /** Sets the position of the body to <code>p</code>. */
  public void setPosition(Point2 p)
    {
    if (getOwner() != null) position.set(getOwner().transpose(p));
    else position.set(p);
    }

  /** 
   *  Sets the velocity of the body. 
   *  @param vx The horizontal velocity.
   *  @param vy The vertical velocity.
   *  @see #setVelocity(Vector2)
   */
  public void setVelocity(double vx, double vy)
    {
    velocity.set(vx, vy);
    }

  /** Sets the velocity of the body to <code>v</code>. */
  public void setVelocity(Vector2 v)
    {
    velocity.set(v);
    }

  /** 
   *  Sets the acceleration of the body. 
   *  @param ax The horizontal acceleration.
   *  @param ay The vertical acceleration.
   *  @see #setAcceleration(Vector2)
   */
  public void setAcceleration(double ax, double ay)
    {
    acceleration.set(ax, ay);
    }

  /** Sets the acceleration of the body to <code>a</code>. */
  public void setAcceleration(Vector2 a)
    {
    acceleration.set(a);
    }

  /**
   *  Advances the time <code>dt</code> seconds and
   *  updates the position and velocity of the body accordingly.
   *  @see #futurePosition(double)
   *  @see #futureVelocity(double)
   */
  public void advance(double dt)
    {
    setPosition(futurePosition(dt));
    setVelocity(futureVelocity(dt));
    // swap the overlap sets and clear the current set
    Set temp = previousOverlaps;
    previousOverlaps = currentOverlaps;
    currentOverlaps = temp;
    currentOverlaps.clear();
    }


	/**
	 *	Returns the position of the body after <time>dt</time> seconds.
   *  The new horizontal position <i>x</i>' is calculated from the equation
   *  <i>x</i>' = <i>a<sub>x</sub></i><i>t</i> + 
   *  <i>v<sub>x</sub></i><i>t</i> + <i>x</i>, and the vertical 
   *  position is calculated accordingly. 
   *  @since 1.1
	 *  @see #advance(double)
	 *	@see #futureVelocity(double)
	 */

	public Point2 futurePosition(double dt)
		{
    double x = position.getX();
    double y = position.getY();
    double vx = velocity.getDeltaX();
    double vy = velocity.getDeltaY();
    double ax = acceleration.getDeltaX();
    double ay = acceleration.getDeltaY();
    double factor = (dt * dt) / 2.0;
    x = (factor * ax) + (vx * dt) + x;
    y = (factor * ay) + (vy * dt) + y;
    return new Point2(x, y);
		}
		

	/**
	 *	Returns the velocity of the body after <time>dt</time> seconds.
	 *  The new horizontal velocity 
   *  <i>v<sub>x</sub></i>' is calculated from the equation
   *  <i>v<sub>x</sub></i>' = 
   *  <i>a<sub>x</sub></i><i>t</i> + <i>v<sub>x</sub></i>, and the vertical
   *  velocity is calculated accordingly.
	 *  @since 1.1
	 *  @see #advance(double)
	 *	@see #futurePosition(double)
	 */

	public Vector2 futureVelocity(double dt)
		{
    double vx = velocity.getDeltaX();
    double vy = velocity.getDeltaY();
    double ax = acceleration.getDeltaX();
    double ay = acceleration.getDeltaY();
    vx = (ax * dt) + vx;
    vy = (ay * dt) + vy;
    return new Vector2(vx, vy);
		}


  /** 
   *  Checks whether the body overlaps the body 
   *  <code>other</code>. An overlap occurs if the distance
   *  between the centers of the bodies is smaller than
   *  the sum of the respective radii. 
   *  @see #isColliding(Body)
   */
  public boolean isOverlapping(Body other)
    {
    double thisX, thisY, otherX, otherY;
    if (getOwner() != null && getOwner().getType() != Universe.INFINITE) 
      {
      thisX = getOwner().getCenter().getX();
      thisY = getOwner().getCenter().getY();
      Vector2 c = position.vectorTo(getOwner().getCenter());
      Point2 p = (Point2)other.getPosition().clone();
      p.transfer(c);
      p = getOwner().transpose(p);
      otherX = p.getX();
      otherY = p.getY();      
      }
    else
      {
      thisX = position.getX();
      thisY = position.getY();
      otherX = other.getPosition().getX();
      otherY = other.getPosition().getY();
      }
    double otherR = other.getRadius();
    // For quick evaluation, compare first the bounding boxes.
    if ((thisX - radius) > (otherX + otherR) ||
        (thisX + radius) < (otherX - otherR) ||
        (thisY - radius) > (otherY + otherR) ||
        (thisY + radius) < (otherY - otherR)) return false;
    // Compare the bounding circles.
    double dx = thisX - otherX;
    double dy = thisY - otherY;
    double squaredDistance = (dx * dx) + (dy * dy);
    double squaredRadiiSum = (radius + otherR) * (radius + otherR);
    if (!(squaredDistance < squaredRadiiSum)) return false;
    currentOverlaps.add(other);
    return true;
    }

  /** 
   *  Returns whether the body collides with the body 
   *  <code>other</code>. A collision occurs when the bodies
   *  overlap each other for the first time (i.e., at the 
   *  moment of impact) and both bodies are corporeal.
   *  @see #isOverlapping(Body)
   *  @see #isCorporeal()
   */
  public boolean isColliding(Body other)
    {
    if (!isCorporeal() || !other.isCorporeal()) return false;
    if (!isOverlapping(other)) return false;
    return !previousOverlaps.contains(other);
    }
  
  /**
   *  Realizes an elastic collision for <em>both</em> the body and
   *  the body <code>other</code>. The linear momentum and kinetic 
   *  energy of the bodies is conserved in the collision.
   *  The new velocity <i>v</i><sub>1</sub>' is calculated from
   *  the equation 
   *  <i>v</i><sub>1</sub>' = 
   *  <i>v</i><sub>1</sub> (<i>m</i><sub>1</sub> - <i>m</i><sub>2</sub>)
   *  / (<i>m</i><sub>1</sub> + <i>m</i><sub>2</sub>) + 
   *  <i>v</i><sub>2</sub> 2<i>m</i><sub>2</sub> / 
   *  (<i>m</i><sub>1</sub> + <i>m</i><sub>2</sub>) 
   *  and <i>v</i><sub>2</sub>' from the equation
   *  <i>v</i><sub>2</sub>' = 
   *  <i>v</i><sub>1</sub> 2<i>m</i><sub>1</sub> /  
   *  (<i>m</i><sub>1</sub> + <i>m</i><sub>2</sub>) +
   *  <i>v</i><sub>2</sub> (<i>m</i><sub>2</sub> - <i>m</i><sub>1</sub>)
   *  / (<i>m</i><sub>2</sub> + <i>m</i><sub>1</sub>).
   *  The new acceleration <i>a</i><sub>1</sub>' is calculated from
   *  from the equation
   *  <i>a</i><sub>1</sub>' = 
   *  <i>m</i><sub>2</sub><i>a</i><sub>2</sub> / <i>m</i><sub>1</sub>,
   *  and the new acceleration <i>a</i><sub>2</sub>' is
   *  calculated accordingly.
   *  The collision is realized regardless whether the bounding circles 
   *  ovelap or not.
   *  @see #inelasticCollision(Body)
   *  @see #isOverlapping(Body)
   */
  public void elasticCollision(Body other)
    {
    Vector2 w = new Vector2(position, other.getPosition());
    double theta = Math.atan2(w.getDeltaY(), w.getDeltaX());
    Vector2 v1 = velocity;
    Vector2 v2 = other.getVelocity();
    v1.rotate(-theta);
    v2.rotate(-theta);
    double m1 = mass;
    double m2 = other.getMass();
    double v1x = v1.getDeltaX() * (m1 - m2) / (m1 + m2) 
                 +  v2.getDeltaX() * (2 * m2) / (m1 + m2);
    double v2x = v1.getDeltaX() * (2 * m1) / (m1 + m2) 
                 + v2.getDeltaX() * (m2 - m1) / (m2 + m1);
    v1.setDeltaX(v1x);
    v2.setDeltaX(v2x);
    v1.rotate(theta);
    v2.rotate(theta);
    setVelocity(v1);
    other.setVelocity(v2);
    // Update accelerations.
    Vector2 a1 = acceleration;
    Vector2 a2 = other.getAcceleration();
    a1.rotate(-theta);
    a2.rotate(-theta);
    double a1x = (m2 * a2.getDeltaX()) / m1;
    double a2x = (m1 * a1.getDeltaX()) / m2;
    a1.setDeltaX(a1x);
    a2.setDeltaX(a2x);
    a1.rotate(theta);
    a2.rotate(theta);
    setAcceleration(a1);
    other.setAcceleration(a2);
    }



  /**
   *  Realizes an inelastic collision for <em>both</em> the body and
   *  the body <code>other</code>. The linear momentum of the bodies is 
   *  conserved but kinetic energy is lost.
   *
   *	@since 1.1
   *  @see #elasticCollision(Body)
   *  @see #isOverlapping(Body)
   */
  public void inelasticCollision(Body other)
    {
    Vector2 w = new Vector2(position, other.getPosition());
    double theta = Math.atan2(w.getDeltaY(), w.getDeltaX());
    Vector2 v1 = velocity;
    Vector2 v2 = other.getVelocity();
    v1.rotate(-theta);
    v2.rotate(-theta);
    double m1 = mass;
    double m2 = other.getMass();
    
    double initialMomentum = (m1 * v1.getDeltaX()) + (m2 * v2.getDeltaX());
    
		double vres = initialMomentum / (m1 + m2);
    
    v1.setDeltaX(vres);
    v2.setDeltaX(vres);
    v1.rotate(theta);
    v2.rotate(theta);
    setVelocity(v1);
    other.setVelocity(v2);
    }


  /** Returns a string that contains the values of the attributes. */
  public String toString()
    {
    return "mass = " + mass + " radius = " + radius + " position = " 
           + position + " velocity = " + velocity + " acceleration = "
           + acceleration + " " + super.toString();
    }
  }