/* 
 * BoundaryLine.java 1.1 2002-06-11
 *
 *  2002 Jouni Smed
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of the University of Turku nor the names of its
 *       contributors may be used to endorse or promote products derived 
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package fi.utu.cs.physics;

/**
 *  A boundary line. 
 * 
 *  @since 1.1
 *  @version 1.1, 020611
 *  @author  Jouni Smed
 */

public class BoundaryLine extends Boundary
  {    
  /** Non-vertical and non-horizontal boundary line. */
  public static final int OTHER = 0;
  /** A horizontal boundary line. */
  public static final int HORIZONTAL = 1;
  /** A vertical boundary line. */
  public static final int VERTICAL   = 2;
  
  /** First end point of the boundary line. */
  private Point2 endPoint1;
  /** Second end point of the boundary line. */
  private Point2 endPoint2;
  /** The type of the boundary line. */
  private int type;
  
  /** 
   *  Constructs a boundary line from point (<code>px1</code>, 
   *  <code>py1</code>) to point (<code>px2</code>, <code>py2</code>). 
   */
  public BoundaryLine(double px1, double py1, double px2, double py2)
  	{
  	this(new Point2(px1, py1), new Point2(px2, py2));
  	}
  
  
  /** 
   *  Constructs a boundary line from point <code>p1</code> to
   *  point <code>p2</code>. 
   */
  public BoundaryLine(Point2 p1, Point2 p2)
    {
    super();
    type = OTHER;
  	if (Point2.HORIZONTAL.compare(p1, p2) == 0) 
  		{
  		type = VERTICAL;
  		// If line is vertical, endPoint1.getY() < endPoint2.getY()
	    endPoint1 = p1.getY() < p2.getY() ? 
	    	(Point2)p1.clone() : (Point2)p2.clone();
	    endPoint2 = p1.getY() < p2.getY() ? 
	    	(Point2)p2.clone() : (Point2)p1.clone();
  		}
  	else if (Point2.VERTICAL.compare(p1, p2) == 0) 
  		{
  		type = HORIZONTAL;
  		// If line is horizontal, endPoint1.getX() < endPoint2.getX()
	    endPoint1 = p1.getX() < p2.getX() ? 
	    	(Point2)p1.clone() : (Point2)p2.clone();
	    endPoint2 = p1.getX() < p2.getX() ? 
	    	(Point2)p2.clone() : (Point2)p1.clone();
  		}
  	else
  		{
	    endPoint1 = (Point2)p1.clone();
	    endPoint2 = (Point2)p2.clone();  		
  		}
    }
  
  /**
   *	Returns a two-element array comprising the end points.
   */
  public Point2[] getEndPoints()
  	{
  	Point2[] result = new Point2[2];
  	result[0] = (Point2)endPoint1.clone();
  	result[1] = (Point2)endPoint2.clone();
  	return result;
  	}
  
  /** 
  	*	Returns the type of the boundary line.
  	*
  	* @see #HORIZONTAL
  	*	@see #VERTICAL
  	*	@see #OTHER
  	*/
  public int getType()
  	{
  	return type;
  	}
 

	/**
	 *	Returns <code>true</code>, if the bounding box of the path 
	 *  (<code>x1</code>, <code>y1</code>)-(<code>x2</code>, <code>y2</code>)
	 *	of a circle with a radius <code>r</code> crosses the boundary line; 
	 *  otherwise, returns <code>false</code>.
	 */
	protected boolean isCrossedByBoundingBox(double x1, double y1,
																					 double x2, double y2, double r)
		{
		double xmin = Math.min(x1 - r, x2 - r);
		double xmax = Math.max(x1 + r, x2 + r);
 		double ymin = Math.min(y1 - r, y2 - r);
		double ymax = Math.max(y1 + r, y2 + r);
		
		if (type == VERTICAL)
			{
			double x = endPoint1.getX();
			double yp1 = endPoint1.getY();
			double yp2 = endPoint2.getY();
			return (xmin <= x & x <= xmax) & (ymin <= yp2 & yp1 <= ymax);
			}
		if (type == HORIZONTAL)
			{
			double y = endPoint1.getY();
			double xp1 = endPoint1.getX();
			double xp2 = endPoint2.getX();
			return (ymin <= y & y <= ymax) & (xmin <= xp2 & xp1 <= xmax);
			}

		double xpmin = Math.min(endPoint1.getX(), endPoint2.getX());
		double xpmax = Math.max(endPoint1.getX(), endPoint2.getX());
		double ypmin = Math.min(endPoint1.getY(), endPoint2.getY());
		double ypmax = Math.max(endPoint1.getY(), endPoint2.getY());

		return (xmin <= xpmax & xpmin <= xmax)
				 & (ymin <= ypmax & ypmin <= ymax);
		}


	/**
	 *	Returns <code>true</code>, if the line (<code>x1</code>,
	 *  <code>y1</code>)-(<code>x2</code>, <code>y2</code>)
	 *	crosses the boundary line; otherwise, returns <code>false</code>.
	 */
	protected boolean isCrossedByLine(double x1, double y1, double x2, double y2)
		{
			if (type == VERTICAL | type == HORIZONTAL)
				return isCrossedX(x1, y1, x2, y2) && isCrossedY(x1, y1, x2, y2);
			
			double x3 = endPoint1.getX();
			double y3 = endPoint1.getY();
			double x4 = endPoint2.getX();
			double y4 = endPoint2.getY();
			
			double d = ((y4 - y3) * (x2 - x1)) - ((x4 - x3) * (y2 - y1));
			// Are the lines parallel?
			if (boundaryEpsilon.compare(d, 0.0) == 0) return false;
			
			double ua = (((x4 - x3) * (y1 - y3)) - ((y4 - y3) * (x1 - x3))) / d;
			double ub = (((x2 - x1) * (y1 - y3)) - ((y2 - y1) * (x1 - x3))) / d;
			
			return (0.0 <= ua & ua <= 1.0) & (0.0 <= ub & ub <= 1.0);
		}


	/**
	 *	Returns <code>true</code>, if the path 
	 *  (<code>x1</code>, <code>y1</code>)-(<code>x2</code>, <code>y2</code>)
	 *	crosses the boundary line vertically; otherwise, returns 
	 *  <code>false</code>.
	 */
  private boolean isCrossedX(double x1, double y1, double x2, double y2)
  	{
  	if (type == HORIZONTAL) 
  		{
  		double deltaY = y2 - y1;
  		double deltaX = x2 - x1;
  		double y = endPoint1.getY() - y1;
  		double x = (boundaryEpsilon.compare(deltaY, 0.0) != 0) ? 
					 			 ((deltaX / deltaY) * y + x1) : x2;
  		return (x >= endPoint1.getX()) &
  		       (x <= endPoint2.getX());
			}  		     
  	if (type == VERTICAL)
  		{
  		double xMin = Math.min(x1, x2);
  		double xMax = Math.max(x1, x2);
  		double x = endPoint1.getX();
  		return ((xMin < x) & (x < xMax));
  		}
  	return false;
  	}


	/**
	 *	Returns <code>true</code>, if the path 
	 *  (<code>x1</code>, <code>y1</code>)-(<code>x2</code>, <code>y2</code>)
	 *	crosses the boundary line horizontally; otherwise, returns 
	 *  <code>false</code>.
	 */
	private boolean isCrossedY(double x1, double y1, double x2, double y2)
		{
  	if (type == VERTICAL) 
  		{
  		double deltaY = y2 - y1;
  		double deltaX = x2 - x1;
			double x = endPoint1.getX() - x1;
  		double y = (boundaryEpsilon.compare(deltaX, 0.0) != 0) ? 
					 			 ((deltaY / deltaX) * x + y1) : y2;
 			return (y >= endPoint1.getY()) &
 			       (y <= endPoint2.getY());
  		}
  	if (type == HORIZONTAL)
  		{
  		double yMin = Math.min(y1, y2);
  		double yMax = Math.max(y1, y2);
  		double y = endPoint1.getY();
  		return ((yMin < y) & (y < yMax));
  		}
		return false;
		}


  /**
   *  Realizes an elastic collision for the static boundary and
   *  the body <code>other</code>.
   *  The linear momentum and kinetic energy of the body is conserved 
   *  in the collision.
   */
  public void elasticCollision(Body other)
    {    
    Vector2 v2 = other.getVelocity();
    Vector2 a2 = other.getAcceleration();
    if (type == VERTICAL)
    	{
    	v2.setDeltaX(-v2.getDeltaX());
    	a2.setDeltaX(-a2.getDeltaX());
    	}
    else if (type == HORIZONTAL)
    	{
    	v2.setDeltaY(-v2.getDeltaY());
    	a2.setDeltaY(-a2.getDeltaY());
    	}
    else
    	{
    	Vector2 tangent = new Vector2(endPoint1, endPoint2);
    	Vector2 normal = (Vector2)tangent.clone();
    	normal.rotate90();
    	Vector2 vt2 = tangent.project(v2);
    	Vector2 vn2 = normal.project(v2);
    	vn2.rotate180();
    	v2 = (Vector2)vt2.clone();
    	v2.add(vn2);
    	Vector2 at2 = tangent.project(a2);
    	Vector2 an2 = normal.project(a2);
    	an2.rotate180();
    	a2 = (Vector2)at2.clone();
    	a2.add(an2);
    	}
		other.setVelocity(v2);		
    other.setAcceleration(a2);
		}


  /**
   *  Realizes an inelastic collision for the static boundary and
   *  the body <code>other</code>.
   *  The linear momentum is conserved but kinetic energy is lost 
   *  in the collision.
   */
  public void inelasticCollision(Body other)
    {    
    Vector2 v2 = other.getVelocity();
    Vector2 a2 = other.getAcceleration();
    if (type == VERTICAL)
    	{
    	v2.setDeltaX(0.0);
    	a2.setDeltaX(0.0);
    	}
    else if (type == HORIZONTAL)
    	{
    	v2.setDeltaY(0.0);
    	a2.setDeltaY(0.0);
    	}
    else
    	{
    	Vector2 tangent = new Vector2(endPoint1, endPoint2);
    	v2 = tangent.project(v2);
    	a2 = tangent.project(a2);
    	}
		other.setVelocity(v2);		
    other.setAcceleration(a2);
		}


  /** Returns a string that contains the values of the attributes. */
  public String toString()
    {
    return "line = " + endPoint1 + "-" + endPoint2 + " " + super.toString();
    }
	}
