/* 
 * MomentaryBody.java 1.1 2002-06-04
 *
 *  2001-2002 Jouni Smed
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of the University of Turku nor the names of its
 *       contributors may be used to endorse or promote products derived 
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package fi.utu.cs.physics;

/**
 *  A moving physical body with mass, dimensionality and a finite lifetime. 
 *  The lifetime is decreased each time the <code>advance</code> method
 *  is called. When lifetime becomes less than zero, the <code>die</code>
 *  method is executed.
 *
 *  @version 1.1, 020604
 *  @author  Jouni Smed
 */

public abstract class MomentaryBody extends Body 
  {
  /** The remaining lifetime of this object. */
  private double lifetime;

  /** 
   *  Constructs a body with the mass <code>m</code>, 
   *  the radius <code>r</code> and the lifetime <code>lt</code>. 
   *  The body resides in the origin and has a zero velocity and 
   *  a zero acceleration. The body is corporeal and has no owner.  
   */
  protected MomentaryBody(double m, double r, double lt)
    {
    this(m, r, lt, Point2.ORIGIN, Vector2.ZERO, Vector2.ZERO);
    }

  /** 
   *  Constructs a body with the mass <code>m</code>,
   *  the radius <code>r</code>, the lifetime <code>lt</code> 
   *  and the position <code>p</code>. 
   *  The body has a zero velocity and a zero acceleration.
   *  The body is corporeal and has no owner.  
   */
  protected MomentaryBody(double m, double r, double lt, Point2 p)
    {
    this(m, r, lt, p, Vector2.ZERO, Vector2.ZERO);
    }
  
  /** 
   *  Constructs a body with the mass <code>m</code>,
   *  the radius <code>r</code>, the lifetime <code>lt</code>,
   *  the position <code>p</code>
   *  and the velocity <code>v</code>.  The body has a 
   *  zero acceleration. The body is corporeal and has no owner.  
   */
  protected MomentaryBody(double m, double r, double lt, Point2 p, Vector2 v)
    {
    this(m, r, lt, p, v, Vector2.ZERO);
    }

  /** 
   *  Constructs a body with the mass <code>m</code>,
   *  the radius <code>r</code>, the lifetime <code>lt</code>,
   *  the position <code>p</code>,
   *  the velocity <code>v</code> and the acceleration <code>a</code>.
   *  The body is corporeal and has no owner.  
   */
  protected MomentaryBody(double m, double r, double lt, 
                          Point2 p, Vector2 v, Vector2 a)
    {
    super(m, r, p, v, a);
    setLifetime(lt);
    }


  /** Returns the remaining lifetime of this body. */
  public double getLifetime()
    {
    return lifetime;
    }

  /** Sets the lifetime to <code>lt</code> seconds. */
  public void setLifetime(double lt)
    {
    lifetime = lt;
    }

  /**
   *  Advances the time <code>dt</code> seconds and
   *  updates the position, velocity and lifetime of the body accordingly.
   *  @see Body#advance(double)
   */
  public void advance(double dt)
    {
    lifetime -= dt;
    if (lifetime < 0.0) die();
    else super.advance(dt);
    }

  /** Realizes the death of this body. */
  public abstract void die();

  /** Returns a string that contains the values of the attributes. */
  public String toString()
    {
    return "lifetime = " + lifetime + " " + super.toString();
    }
  }
  