/* 
 * Universe.java 1.1 2002-06-11
 *
 *  2001-2002 Jouni Smed
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of the University of Turku nor the names of its
 *       contributors may be used to endorse or promote products derived 
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package fi.utu.cs.physics;

import java.util.List;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
import java.util.Collections;
import java.util.Observable;

/**
 *  A universe comprising a collection of <code>Body</code> objects.
 *  The type of the universe can be infinite, finite or toroid. 
 *  Classes implementing the <code>Observer</code> interface can 
 *  observe the universe for detecting collisions between the corporeal bodies;
 *  an instance of the <code>CollisionPair</code> inner class is
 *  submitted to the observers.
 *
 *  @see Particle
 *  @see Body
 *	@see Boundary
 *  @version 1.1, 020611
 *  @author  Jouni Smed
 */

public class Universe extends Observable
  {
  /** 
   *  An infinite universe which does not impose any limits to the
   *  position of the bodies. 
   */
  public static final int INFINITE = 0;
  /** 
   *  A finite universe which limits the position of the bodies.
   *  The shape of the universe is the rectangle 
   *  (<code>getMinimumCorner()</code>, <code>getMaximumCorner()</code>). 
   *  @see #getMinimumCorner()
   *  @see #getMaximumCorner()
   */
  public static final int FINITE = 1;
  /** 
   *  An infinite universe of finite size where the opposite
   *  vertical and horizontal edges are connected to form a torus.
   *  The shape of the universe is the rectangle 
   *  (<code>getMinimumCorner()</code>, <code>getMaximumCorner()</code>). 
   *  @see #getMinimumCorner()
   *  @see #getMaximumCorner()
   */
  public static final int TOROID = 2;  
  /** 
   *  A linked list which collects the bodies. The list is partitioned
   *  so that the ethereal bodies are in the beginning and the corporeal
   *  bodies in the end of the list.
   */
  private LinkedList bodyList;
  /**
   *  A linked list which collects the boundaries. 
   *	@since 1.1
   */
  private LinkedList boundaryList;
  /** The type of the universe. */
  private int type;
  /** 
   *  The minimum (i.e., lower left) corner of the universe if it is 
   *  finite or toroid; <code>null</code> if the universe is infinite.
   */
  private Point2 minimumCorner;
  /** 
   *  The maximum (i.e., upper right corner of the universe if it is 
   *  finite or toroid; <code>null</code> if the universe is infinite.
   */
  private Point2 maximumCorner;
  /** 
   *  The center of the universe if it is finite or toroid;
   *  <code>null</code> if the universe is infinite.
   */
  private Point2 center;

  /** A monitor indicating whether the collection can be updated. */
  private boolean noUpdatesOnBodyList;
  /** A queue for insertion operations. */
  private LinkedList insertionQueue;  
  /** A queue for removal operations. */
  private LinkedList removalQueue;
  
  /**
   *	Comprises the the (body, particle) pairs that collide
   *	inelastically.
   *	@since 1.1
   */
  private Map inelasticColliders;
  
	/** 
	 *  Acceleration of gravity, which is used when calculating frictions.
	 *	@since 1.1
	 */
	private double accelerationOfGravity = 9.80665;

  /**
   *  Comprises a <code>Body</code> object and a <code>Particle</code>
   *  object that have been the counterparts in a collision. The 
   *  <code>Universe</code> class, which is inherited from 
   *  <code>Observable</code>, submits an instance of this class to 
   *  its observers. The notification is sent if and only if
   *  (1) a body and a particle collide, and (2) either the body or
   *  the particle wants its collisions to be notified.
   *
   *  @see Particle#addCollisionNotification(Class)
   *  @see java.util.Observable#notifyObservers(Object)
   *  @see java.util.Observer#update(Observable,Object)
   */
  public class CollisionNotification
    {
    /** None of the counterparts. */
    public static final int NONE   = 0;
    /** 
     *  The first counterpart. Always an instance of the class
     *  <code>Body</code>.
     */
    public static final int FIRST  = 1;
    /** The second counterpart. Always an instance of the class
     *  <code>Particle</code>.
     */
    public static final int SECOND = 2;
    /** Both of the counterparts. */
    public static final int BOTH   = 3;
    
    /** The first counterpart. */
    private Body body;
    /** The second counterpart. */
    private Particle particle;

    /** 
     *  Creates a notification for the bodies <code>b1</code> 
     *  and <code>b2</code> 
     */
    public CollisionNotification(Body b1, Body b2) 
      {
      body = b1;
      particle = b2;
      }
      
    /** 
     *  Creates a notification for the body <code>b</code> 
     *  and the particle <code>p</code> 
     *  @since 1.1
     */
    public CollisionNotification(Body b, Particle p) 
      {
      body = b;
      particle = p;
      }

      
    /** 
     *  Returns the <code>i</code>th counterpart of the collision.
     *  Note: Throws a <code>ClassCastException</code> if 
     *  <code>i == SECOND</code> and the second counterpart is not an 
     *  instance of the class <code>Body</code>.
	   * 	@deprecated  As of version 1.1, replaced by {@link #getParticle(int)}.
     */
    public Body getCounterpart(int i)
      {
      return (Body)getParticle(i);
      }

    
    /**
     *	Returns the body involved in the collision.
     *  @since 1.1
     */
    public Body getBody()
    	{
    	return body;
    	}
    
    /**
     *	Returns the particle involved in the collision.
     *  @since 1.1
     */
    public Particle getParticle()
    	{
    	return particle;
    	}

    /** 
     *  Returns the <code>i</code>th counterpart of the collision.
     *  Returns <code>null</code>, if <code>i != FIRST</code> and
     *  <code>i != SECOND</code>.
     *  @since 1.1
     *  @see #FIRST
     *  @see #SECOND
     */
    public Particle getParticle(int i)
    	{
      return (i == FIRST ? body : (i == SECOND ? particle : null));
    	}
    
    /** 
     *  Returns which of the counterparts is of the type <code>t</code>. 
     *  @see #NONE
     *  @see #FIRST
     *  @see #SECOND
     *  @see #BOTH
     */
    public int counterpart(Class t)
      {
      int result = NONE;
      if (body.getClass().equals(t)) result |= FIRST;
      if (particle.getClass().equals(t)) result |= SECOND;
      return result;
      }
    }
  
  /**
   *  Creates an infinite universe. After the creation, the universe
   *  is devoid of bodies.
   */
  public Universe()
    {
    this(INFINITE, null, null);
    }

  /**
   *  Creates a universe. After the creation, the universe is devoid
   *  of bodies and boundaries. The collisions are set fully elastic.
   *  @param t The type of the universe (infinite, finite or toroid).
   *  @param min The minimum corner of the universe.
   *  @param max The maximum corner of the universe.
   *  @see #INFINITE
   *  @see #FINITE
   *  @see #TOROID
   */
  public Universe(int t, Point2 min, Point2 max)
    {
    bodyList = new LinkedList();
    boundaryList = new LinkedList();
    type = t;
    minimumCorner = min;
    maximumCorner = max;
    if (min != null && max != null)
      {
      double cx = (max.getX() - min.getX()) / 2 + min.getX();
      double cy = (max.getY() - min.getY()) / 2 + min.getY();
      center = new Point2(cx, cy);
      }
    else center = null;
    noUpdatesOnBodyList = false;
    insertionQueue = new LinkedList();
    removalQueue = new LinkedList();
    inelasticColliders = new HashMap();
    }

  /** 
   *  Returns the type of the universe. 
   *  @see #INFINITE
   *  @see #FINITE
   *  @see #TOROID
   */
  public int getType()
    {
    return type;
    }

  /** 
   *  Returns the minimum corner of the universe or <code>null</code>
   *  if the universe is infinite.
   */
  public Point2 getMinimumCorner()
    {
    return (Point2)minimumCorner.clone();
    }

  /** 
   *  Returns the maximum corner of the universe or <code>null</code>
   *  if the universe is infinite.
   */
  public Point2 getMaximumCorner()
    {
    return (Point2)maximumCorner.clone();
    }

  /** 
   *  Returns the center of the universe or <code>null</code>
   *  if the universe is infinite.
   */
  public Point2 getCenter()
    {
    return (Point2)center.clone();
    }


	/**
	 *	Adds the class pair (<code>c1</code>, 
	 *  <code>c2</code>) to the inelastic colliders.
	 *	@since 1.1
	 */
	private void addInelasticPair(Class c1, Class c2)
		{
		Set s = (Set)inelasticColliders.get(c1);
		if (s == null)
			{
			s = new HashSet();
			inelasticColliders.put(c1, s);
			}
		s.add(c2);
		}


	/**
	 *	Removes the class pair (<code>c1</code>, 
	 *  <code>c2</code>) from the inelastic colliders.
	 *	@since 1.1
	 */
	private void removeInelasticPair(Class c1, Class c2)
		{
		Set s = (Set)inelasticColliders.get(c1);
		if (s != null)
			{
			s.remove(c2);
			if (s.isEmpty()) inelasticColliders.remove(c1);
			}
		}


	/**
	 *	Returns <code>true</code> if collisions between instances of the classes
	 *  <code>bc</code> and <code>pc</code> are inelastic; otherwise,
	 *  returns <code>false</code>. 
	 *
	 *	@since 1.1
	 *	@see #setInelasticCollision(Class,Class)
	 *	@see #clearInelasticCollision(Class,Class)
	 */
	public boolean isInelasticCollision(Class bc, Class pc)
		{
		Set s = (Set)inelasticColliders.get(bc);
		if (s == null) return false;
		return s.contains(pc);
		}

	/**
	 *	Returns <code>true</code> if a collision between the body
	 *  <code>b</code> and the particle <code>p</code> is inelastic; otherwise,
	 *  returns <code>false</code>. 
	 *
	 *	@since 1.1
	 */
	public boolean isInelasticCollision(Body b, Particle p)
		{
		return isInelasticCollision(b.getClass(), p.getClass());
		}


	/**
	 *	Sets the collisions between instances of the classes
	 *  <code>bc</code> and <code>pc</code> to be inelastic.
	 *
	 *	@since 1.1
	 *	@see #isInelasticCollision(Class,Class)
	 *	@see #clearInelasticCollision(Class,Class)
	 */
	public void setInelasticCollision(Class bc, Class pc)
		{
		addInelasticPair(bc, pc);
		addInelasticPair(pc, bc);
		}

	/**
	 *	Sets the collisions between instances of the classes
	 *  <code>bc</code> and <code>pc</code> to be elastic.
   *
	 *	@since 1.1
	 *	@see #isInelasticCollision(Class,Class)
	 *	@see #setInelasticCollision(Class,Class)
	 */
	public void clearInelasticCollision(Class bc, Class pc)
		{
		removeInelasticPair(bc, pc);
		removeInelasticPair(pc, bc);
		}


  /** 
   *  Adds the body <code>b</code> to the universe, which is also set
   *  to be the owner of the body.
   *  @see Body#setOwner(Universe)
   */
  public void addBody(Body b)
    {
    if (noUpdatesOnBodyList) insertionQueue.addLast(b);
    else realizeAddBody(b);
    }

  /** Does the actual insertion. */
  private void realizeAddBody(Body b)
    {
    b.setOwner(this);
    if (b.isCorporeal()) 
    	{
    	bodyList.addLast(b);
    	}
    else 
    	{
    	bodyList.addFirst(b);
    	}
    }

  /**
   *  Removes the body <code>b</code> from the universe and sets its
   *  owner to <code>null</code>.
   *  @see Body#setOwner(Universe)
   */
  public void removeBody(Body b)    
    {
    if (noUpdatesOnBodyList) removalQueue.addLast(b);
    else realizeRemoveBody(b);
    }

  /** Does the actual removal. */
  private void realizeRemoveBody(Body b)
    {
    bodyList.remove(b);
    b.setOwner(null);
    }

  /** 
   *  Adds the boundary <code>b</code> to the universe, which is also set
   *  to be the owner of the body.
   *  @since 1.1
   *  @see Boundary#setOwner(Universe)
   */
  public void addBoundary(Boundary b)
    {
    if (noUpdatesOnBodyList) insertionQueue.addLast(b);
    else realizeAddBoundary(b);
    }

  /** 
   *	Does the actual insertion. 
   *  @since 1.1
   */
  private void realizeAddBoundary(Boundary b)
    {
    b.setOwner(this);
    boundaryList.addFirst(b);
    }

  /**
   *  Removes the boundary <code>b</code> from the universe and sets its
   *  owner to <code>null</code>.
   *  @since 1.1
   *  @see Boundary#setOwner(Universe)
   */
  public void removeBoundary(Boundary b)    
    {
    if (noUpdatesOnBodyList) removalQueue.addLast(b);
    else realizeRemoveBoundary(b);
    }

  /** 
   *	Does the actual removal. 
   *  @since 1.1
   */
  private void realizeRemoveBoundary(Boundary b)
    {
    boundaryList.remove(b);
    b.setOwner(null);
    }

  /**
   *  Removes all bodies and boundaries from the universe and sets their
   *  owner to <code>null</code>.
   *  @see Body#setOwner(Universe)
   *  @see Boundary#setOwner(Universe)
   */
  public void removeAll()
    {
    if (noUpdatesOnBodyList)
    	{
    	ListIterator it = bodyList.listIterator();
    	while (it.hasNext()) removalQueue.addLast(it.next());
    	it = boundaryList.listIterator();
    	while (it.hasNext()) removalQueue.addLast(it.next());
    	}
    else 
    	{
    	bodyList.clear();
    	boundaryList.clear();
			}
    }

  /** Executes the queued insertions and removals. */
  private void updateBodyList()
    {
    if (noUpdatesOnBodyList) return;
    if (!insertionQueue.isEmpty())
      {
      ListIterator it = insertionQueue.listIterator();
      while (it.hasNext())
      	{
      	Particle p = (Particle)it.next();
      	if (p instanceof Body) realizeAddBody((Body)p);
      	else if (p instanceof Boundary) realizeAddBoundary((Boundary)p);
      	else throw new RuntimeException("Universe: Tried to insert an unrecognized particle type.");
      	}
      insertionQueue.clear();
      }
    if (!removalQueue.isEmpty())
      {
      ListIterator it = removalQueue.listIterator();
      while (it.hasNext())       
      	{
      	Particle p = (Particle)it.next();
      	if (p instanceof Body) realizeRemoveBody((Body)p);
      	else if (p instanceof Boundary) realizeRemoveBoundary((Boundary)p);
      	else throw new RuntimeException("Universe: Tried to remove an unrecognized particle type.");
				}
      removalQueue.clear();
      }
    }

  /**
   *  Returns an unmodifiable (i.e., "read-only") list of the bodies in 
   *  the universe. Attempts to modify the returned list, whether direct 
   *  or via its iterator, result in an
   *  <code>UnsupportedOperationException</code>.
   *	@see #boundaries()
   *  @see java.util.Collections#unmodifiableList(List)
   */
  public List bodies()
    {
    return Collections.unmodifiableList(bodyList);
    }

  /** 
   *  Returns the number of bodies in the universe. 
   *	@since 1.1
   */
  public int bodySize()
    {
    return bodyList.size();
    }

  /**
   *  Returns the number of bodies of the type <code>t</code>
   *  in the universe. 
   *	@since 1.1
   */
  public int bodySize(Class t)
    {
    int result = 0;
    ListIterator it = bodyList.listIterator();
    while (it.hasNext())
      {
      Body body = (Body)it.next();
      if (body.getClass().equals(t)) result++;
      }
    return result;
    }

  /**
   *  Returns an unmodifiable (i.e., "read-only") list of the boundaries in 
   *  the universe. Attempts to modify the returned list, whether direct 
   *  or via its iterator, result in an
   *  <code>UnsupportedOperationException</code>.
   *  @since 1.1
   *  @see #bodies()
   *  @see java.util.Collections#unmodifiableList(List)
   */
  public List boundaries()
    {
    return Collections.unmodifiableList(boundaryList);
    }

  /** 
   *  Returns the number of boundaries in the universe. 
   *	@since 1.1
   */
  public int boundarySize()
    {
    return boundaryList.size();
    }

  /**
   *  Returns the number of boundaries of the type <code>t</code>
   *  in the universe. 
   *	@since 1.1
   */
  public int boundarySize(Class t)
    {
    int result = 0;
    ListIterator it = boundaryList.listIterator();
    while (it.hasNext())
      {
      Boundary boundary = (Boundary)it.next();
      if (boundary.getClass().equals(t)) result++;
      }
    return result;
    }

  /** 
   *  Returns the number of bodies in the universe. 
   * 	@deprecated  As of version 1.1, replaced by {@link #bodySize()}.
   */
  public int size()
    {
    return bodySize();
    }

  /**
   *  Returns the number of bodies of the type <code>t</code>
   *  in the universe. 
   * 	@deprecated  As of version 1.1, replaced by {@link #bodySize(Class)}.
   */
  public int size(Class t)
    {
    return bodySize(t);
    }
  
  /**
   *  Advances the time <code>dt</code> seconds and
   *  updates its bodies accordingly.
   *  The methods advances the bodies iteratively. After each
   *  advancement, it examines whether the body in question is corporeal 
   *  and does it collide with any of the remaining (i.e., not yet 
   *  advanced) corporeal bodies or boundaries. If a collision occurs, 
   *  the method realizes an elastic collision between the bodies.
   *  If either particle have requested a collision notification on
   *  the other particle, all observers will be notified. 
   *  @see Body#advance(double)
   *  @see Body#isCorporeal()
   *  @see Body#isColliding(Body)
   *  @see Body#elasticCollision(Body)
   *  @see Universe.CollisionNotification
   */
  public void advance(double dt)
    {
    noUpdatesOnBodyList = true;
    ListIterator it = bodyList.listIterator();
    while (it.hasNext())
      {
      Body body = (Body)it.next();
      
      // First, collisions with boundaries.
      ListIterator bounds = boundaryList.listIterator();
      while (bounds.hasNext())
      	{
      	Boundary bound = (Boundary)bounds.next();
      	
      	boolean requiresNotification = body.isCollisionNotified(bound) 
      															 | bound.isCollisionNotified(body);
				boolean isCorporealCollision = body.isCorporeal() 
																		 & bound.isCorporeal();
      	// Do we really have to calculate the path?
      	if (requiresNotification | isCorporealCollision)
      	  {
      	  boolean crossesTheBoundary = bound.isCrossed(body, dt);
      	  if (crossesTheBoundary & requiresNotification)
      	  	{
      	  	setChanged();
      	  	notifyObservers(new CollisionNotification(body, bound));
      	  	}
      	  if (crossesTheBoundary & isCorporealCollision)
      	  	{
      	  	if (isInelasticCollision(body, bound))
							bound.inelasticCollision(body);
						else
	      			bound.elasticCollision(body);
	      		}
	     	  }
      	}      

			// Second, move the body.
      body.advance(dt);
      
      // Third, collisions with other bodies.
      // NB. Only in the latter part of the list needs to be detected.
      ListIterator corps = bodyList.listIterator(it.nextIndex());
      while (corps.hasNext())
        {
        Body otherBody = (Body)corps.next();
        
      	boolean requiresNotification = body.isCollisionNotified(otherBody) 
      															 | otherBody.isCollisionNotified(body);
				boolean isCorporealCollision = body.isCorporeal() 
																		 & otherBody.isCorporeal();
      	// Do we really have to determine the collision?
      	if (requiresNotification | isCorporealCollision)
					{        
					boolean bodiesCollide = body.isColliding(otherBody);
					if (bodiesCollide & requiresNotification)
						{
	          setChanged();
	          notifyObservers(new CollisionNotification(body, otherBody));
	          }
					if (bodiesCollide & isCorporealCollision)
						{
      	  	if (isInelasticCollision(body, otherBody))
							body.inelasticCollision(otherBody);
						else
	      			body.elasticCollision(otherBody);
	          }
         	}
      	}
    	}
    noUpdatesOnBodyList = false;
    updateBodyList();
    }


  /**
   *  Transposes the point <code>p</code> according to the type
   *  of the universe. If the universe is infinite, the method returns 
   *  <code>p</code>. Otherwise, the method returns a new point where 
   *  <code>getMinimumCorner.getX() <= result.getX() <= 
   *  getMaximumCorner.getX()</code> and <code>getMinimumCorner.getY() 
   *  <= result.getY() <= getMaximumCorner.getY()</code>.
   *  If the universe is finite, an exceeding point is moved back inside 
   *  the rectangle.  If the universe is toroid, an exceeding point is 
   *  transposed back inside the rectangle by regarding it as a torus
   *  surface.
   *  @see #getType()
   *  @see #getMinimumCorner()
   *  @see #getMaximumCorner()
   */
  public Point2 transpose(Point2 p)
    {
    if (type == INFINITE) return p;
    Point2 result = (Point2)p.clone();
    if (Point2.HORIZONTAL.compare(minimumCorner, p) > 0) // west?
      {
      switch (type)
        {
        case FINITE: 
            result.setX(minimumCorner.getX());
            break;
        case TOROID: 
            result.add(maximumCorner.getX() - minimumCorner.getX(), 0.0);
            break;
        }
      }
    else if (Point2.HORIZONTAL.compare(maximumCorner, p) < 0) // east?
      {
      switch (type)
        {
        case FINITE: 
            result.setX(maximumCorner.getX());
            break;
        case TOROID: 
            result.add(minimumCorner.getX() - maximumCorner.getX(), 0.0);
            break;
        }       
      }
    if (Point2.VERTICAL.compare(minimumCorner, p) > 0) // south?
      {
      switch (type)
        {
        case FINITE: 
            result.setY(minimumCorner.getY());
            break;
        case TOROID: 
            result.add(0.0, maximumCorner.getY() - minimumCorner.getY());
            break;
        }
      }
    else if (Point2.VERTICAL.compare(maximumCorner, p) < 0) // north?
      {
      switch (type)
        {
        case FINITE: 
            result.setY(maximumCorner.getY());
            break;
        case TOROID: 
            result.add(0.0, minimumCorner.getY() - maximumCorner.getY());
            break;
        }       
      }
    return result;
    }
  
  /**
   *  Returns the acceleration of gravity of this universe.
   *	By default <i>g</i> = 9.80665 m/s.
   *  @since 1.1
   *	@see #setAccelerationOfGravity(double)
   *  @see OrientedBody
   */
  public double getAccelerationOfGravity()
  	{
  	return accelerationOfGravity;
  	}

  /**
   *  Sets the acceleration of gravity of this universe to <code>g</code>.
   *  @since 1.1
   *	@see #getAccelerationOfGravity()
   *  @see OrientedBody
   */
  public void setAccelerationOfGravity(double g)
  	{
  	accelerationOfGravity = g;
  	}
   
  }